shape = {}
# Format for comments. Use {text} to denote where the comment text should go
comment_format = "/* {text} */\n"

class helper(object):
    """ 
    The Helper class stores the methods specific to a diagram language.
    It's passed to the FlowchartBase object to be used when rendering code
    """

    # File extension
    extension = ""

    # Name of helper
    name = ""

    @staticmethod
    def remove(d, k):
        """ Remove a key from dict d """
        if k in d:
            d.pop(k) 

    @staticmethod
    def shape(in_shape, **kwargs):
        """ Translate a shape to this one recognized by the helper """
        return shape.get(in_shape, in_shape) 

    @staticmethod
    def node(id, label=None, **kwargs):
        pass

    @staticmethod
    def edge(fromId, toId, label=None, **kwargs):
        pass

    @staticmethod
    def startDiagram(*args):
        """ Called first during the render to initialise a diagram """
        pass

    @staticmethod
    def endDiagram(*args):
        """ Called last uring the render to finalise a diagram """
        pass


    @staticmethod
    def startSubdiagram(id, **kwargs):
        pass

    @staticmethod
    def endSubdiagram(id, **kwargs):
        pass

    @staticmethod
    def comment(text):
        return comment_format.format(text=text)


from pydiagrams.AWS import AWSContext as Context

# Need to use a helper that can understand AWS syntax, eg: PUML_AWS
from pydiagrams.helpers.PUML_AWS import Helper
# from pydiagrams.helpers.Graphviz_AWS import Helper

with Context(Helper) as c:
    # Set up the components, showing the AWS items within a frame
    app = c.Mobileclient("App")
    
    with c.Frame("AWS") as aws:
        gw = c.AmazonAPIGateway("API Gateway")
        f1 = c.AWSLambda("Function 1")
        f2 = c.AWSLambda("Function 2")
        f3 = c.AWSLambda("Function 3")
        db = c.AmazonDynamoDB("Dynamo DB")

    # App connects to API gateway
    app >> gw

    # The gateway is a proxy for the lambda function
    gw >> [f1, f2, f3]

    #Some connections between functions
    f1 >= f2 % 'Invokes'
    f3 <= f2 % 'Invokes'

    # Function 2 operates on the database
    f2 | db % "Updates"


from . import helper

helper.shape = {
        # Components
        'Cloud':'cloud',
        'Database':'database',
        'Folder':'folder',
        'Frame':'tab',
        'Node':'box3d',
        'Package':'package',
        'Rectangle':'rect',    
        'Component':'component',
        'Interface':'circle',
        'Diamond':'diamond',

        # Views
        'Table'         :   'rect',
        'View'          :   'parallelogram',
        'Package'       :   'invhouse',
        'File'          :   'folder',
        'Integration'   :   'invtrapezium',
        'System'        :   'ellipse',
        'Task'          :   'cds',

        #Architecture
        'Entity'        :   'ellipse',
        'Module'        :   'component',
        'SystemIntegration' : 'rect'
    }

helper.comment_format = "# {text}\n"

class GV_Attributes(dict):
    def __init__(self, d, permitted):
        super().__init__(d)
        self.permitted = permitted

    def __str__(self):
        return '; '.join(['{}="{}"'.format(k,v) for (k,v) in list(self.items()) if v is not None and k in self.permitted]) \

class GV_Item:
    permitted_attributes = "_background area bb bgcolor center charset clusterrank color colorscheme comment compound concentrate Damping defaultdist dim dimen diredgeconstraints distortion dpi epsilon esep fillcolor fixedsize fontcolor fontname fontnames fontpath fontsize forcelabels gradientangle group height href id image imagepath imagepos imagescale inputscale K label label_scheme labeljust labelloc landscape layer layerlistsep layers layerselect layersep layout levels levelsgap lheight lp lwidth margin maxiter mclimit mindist mode model mosek newrank nodesep nojustify normalize notranslate nslimit  ordering orientation orientation outputorder overlap overlap_scaling overlap_shrink pack packmode pad page pagedir pencolor penwidth peripheries pin pos quadtree quantum rankdir ranksep ratio rects regular remincross repulsiveforce resolution root rotate rotation samplepoints scale searchsize sep shape shapefile showboxes sides size skew smoothing sortv splines start style stylesheet target tooltip truecolor URL vertices viewport voro_margin width xdotversion xlabel xlp z".split()
    def __init__(self, id=None, **attrs):
        self.id = id
        self.attrs = GV_Attributes(attrs, GV_Item.permitted_attributes)

    @property
    def Id(self):
        return self.id

    @Id.setter
    def Id(self, value):
        self.id = value
   
    def __str__(self):
        return f'\n{self.Id} [{self.attrs}];'

class GV_Edge:
    # Attributes that are allowed for an Edge (from https://graphviz.gitlab.io/_pages/doc/info/attrs.html)
    permitted_attributes='arrowhead arrowsize arrowtail color colorscheme comment constraint decorate dir edgehref edgetarget edgetooltip edgeURL fillcolor fontcolor fontname fontsize head_lp headclip headhref headlabel headport headtarget headtooltip headURL href id label labelangle labeldistance labelfloat labelfontcolor labelfontname labelfontsize labelhref labeltarget labeltooltip labelURL layer len lhead lp ltail minlen nojustify penwidth pos samehead sametail showboxes style tail_lp tailclip tailhref taillabel tailport tailtarget tailtooltip tailURL target tooltip URL weight xlabel xlp'.split()
    def __init__(self, fromId, toId, **attrs):
        self.fromId = fromId
        self.toId = toId
        self.attrs = GV_Attributes(attrs, GV_Edge.permitted_attributes)

    def __str__(self):
        return '{} -> {} [{}];'.format(self.fromId, self.toId, self.attrs)

class Helper(helper.helper):

    extension   = 'gv'    
    name = 'Graphviz'

    @staticmethod
    def render_attrs(label, attrs, enclosed=True):
        attrs.update({'label':label or ""})
        return \
            ('[' if enclosed else '') \
            + '; '.join(['{}="{}"'.format(k,v) for (k,v) in list(attrs.items())]) \
            + (']' if enclosed else '') \
            + ';' 

    @staticmethod
    def node(id, label, **kwargs):
        shape = 'shape'
        if shape in kwargs:
            if kwargs[shape] == 'Interface':
                kwargs['xlabel'] = label
                label = ""
                kwargs['width']=0.3
            kwargs[shape] = Helper.shape(kwargs[shape])

        return str(GV_Item(id, label=label, **kwargs))

    @staticmethod    
    def edge(fromId, toId, label=None, **kwargs):
        # return '{} -> {}'.format(fromId, toId) + ' ' + Helper.render_attrs(label, kwargs)
        return str(GV_Edge(fromId, toId, label=label, **kwargs))

    note_node_attrs = {
            'shape':'note',
            'fillcolor':'lightyellow',
            'color':'darkred',
            'fontname':'Cambria',
            'fontcolor':'darkred' 
            }

    note_edge_attrs = {
            'color':'darkred',
            'style':'dashed',
            'constraint':'false',
            'arrowhead':'open'
    }

    comment_format = "' {text}\n"

    FONTNAME='Calibri'

    # Graphviz defaults for each type
    graph_defaults = {
        'fontname':FONTNAME,
        'fontsize':20,
        'labelloc':'t'
    }

    node_defaults = {
        'style':'filled',
        'colorscheme':'paired12',
        'fontname':FONTNAME,
        'fontsize':14
        }

    edge_defaults = {
        'fontname':FONTNAME,
        'fontsize':14
    }

    @staticmethod
    def startDiagram(*args, **attrs):
        return """
digraph g {{
#######################################
# base formatting

fontname="{fontname}";
compound=true; #enables edges between clusters

{graph_defaults}

{node_defaults}

{edge_defaults}

#######################################
{attrs}
        """.format(
            fontname='Calibri', 
            graph_defaults = GV_Item('graph', **Helper.graph_defaults),
            node_defaults  = GV_Item('node', **Helper.node_defaults),
            edge_defaults  = GV_Item('edge', **Helper.edge_defaults),
            attrs=Helper.render_attrs(args[1], attrs, False))

    @staticmethod
    def endDiagram(*args):
        return '}'

    cluster_counter = 0

    @staticmethod
    def startSubdiagram(id, label, **kwargs):
        if 'fillcolor' in kwargs:
            kwargs.setdefault('bgcolor', kwargs['fillcolor'])

        cluster_id = id or Helper.cluster_counter

        kwargs.setdefault('label',label)
        attrs = "; ".join(['{}="{}"'.format(k,v) for (k,v) in list(kwargs.items())]) +";"



        Helper.cluster_counter += 1        

        return """
subgraph cluster_{id} {{
    fontsize=18;
{attrs}""".format(id=cluster_id, attrs=attrs)


    @staticmethod
    def endSubdiagram(id, label, **kwargs):
        return "\n}} \n".format(id=id)


# Demonstration of pydigrams, using the "Views" diagram style

# 1: Import the library to draw the type of diagram (Views)
from pydiagrams.Views  import ViewContext

# 2: Import the Helper that draws the diagram in a certain graph language (PUML or Graphviz)
from pydiagrams.helpers.Graphviz import Helper

# 3: Import all the constants (colour names, attribute names) 
from pydiagrams.helpers.constants import *

# Set up a Diagram using 'd' as the global namespace
with ViewContext(Helper, label="Data Architecture") as d:

    # Create a subgraph 'rms'
    with d.Cluster("RMS") as rms:
        # Allocate objects to rms
        saC     = rms.Table('sa_customer')
        s       = rms.Table('store')
        saTH    = rms.Table('sa_tran_head')
        saSD    = rms.Table('sa_store_day')
        saTI    = rms.Table('sa_tran_item')
        sul     = rms.Table('cr_store_uda_lovs_v')

        if Helper.name == 'Graphviz':
            rms.passthrough('{rank=min s saTH saC saSD saTI sul} #Align')
        
        with rms.Cluster("RMS: INFOCRM", style='filled', fillcolor=grey2) as inf:
            inf.node_attrs = {'fixedsize':'true', 'width':'3.5'}

            inf.node_attrs.update(fillcolor=color1)
            vgaF    = inf.View('v_google_ads_Feed')
            vgaLRH  = inf.View('v_google_ads_loyalty_rate_hdr')

            vgaLR   = inf.View('v_google_ads_loyalty_rate *', fillcolor=color2)
            vgaC    = inf.View('v_google_ads_Customers *',  fillcolor=color3)

            vCI     = inf.View('v_customers_in', fillcolor=color4)

            vgaT    = inf.View('v_google_ads_Trans *')
            vgaTH   = inf.View('v_google_ads_Tran_Headers *')

            vgaS    = inf.View('v_google_ads_Stores *', fillcolor=color7)
            vgaD    = inf.View('v_google_ads_dates *', fillcolor=color9)

            sth     = inf.Table('sales_trans_head', fillcolor=color6)
            stl     = inf.Table('sales_trans_lines', fillcolor=color6)

            gabr    = inf.Table('cr_google_ads_brand_regions *', fillcolor=color8)
            gbp     = inf.Table('cr_google_batch_process', fillcolor=color10, fontcolor='white')

            # Relationships

            #vgaF.Sources = [vgaC, vgaT]
            # << = 'Sources'
            vgaF << [vgaC, vgaT]  # translation: Data for vgaF 'comes from' vgaC, vgaT
            vCI >> vgaC           # translation: Data in vCI 'goes to' vgaC

            vgaTH << [sth, vgaS, vgaD]

            vgaS << gabr
            gbp >> vgaD

            vgaT  << [vgaTH, stl]

            vgaLR >> vgaLRH
        
        vgaLR << [saTH, saC, saSD, sul, vgaD, vgaS]

        saC >> vCI
        s   >> vgaS

        rms.edge_attrs = {color:'red', fontcolor:'red', label:'Exported to'}
        saTH >> sth
        saTI >> stl

    with d.Cluster('Informatica', fillcolor=grey1, style='filled') as infm:
        sch =  infm.Task('ST')

        with infm.Cluster("Taskflow: GoogleAdwords.<brand>", style='filled', fillcolor=grey2) as tf:
            tf.node_attrs.update(width='3.0')

            trx         = tf.Task('Task: adwords_<brand>_body')
            trx_header  = tf.Task('Task: adwords_<brand>_head')

            trx >> trx_header

        with infm.Cluster("adwords.bat", fillcolor=grey3, style='filled') as aw:
            fc = aw.Task('File Concatenation')
            fu = aw.Task('File Upload: gsutil')
            fc >> fu

        inf_schtask = infm.Task("Scheduled Task:\\nAdwords <brand>")

        sch >> tf # link a Task item to a Subgraph

        inf_schtask >> aw

        f  = infm.File('Transaction File')
        h  = infm.File('Header File')
        gf = infm.File('Google File')

        infm.edge_attrs = {color:'blue', fontcolor:'blue', label:'Read By'}
        fc << [f, h]

        tf.edge_attrs= {color:'blue', fontcolor:'blue', label:'Creates'}
        trx >> f 
        trx_header >> h

        #infm.aw.edge_attrs= {color:'blue', fontcolor:'blue', label:'Produces'}
        fc >> gf % 'Produces'

        gf >> fu        

    with d.Cluster('Google Cloud Platform') as gcp:
        gcp.node_attrs = {fillcolor:'11'}

        with gcp.Cluster('Storage', fillcolor=grey2, style='filled') as gcst:
            gcs = gcst.File('crg-ads-uploads/<brand>')

        with gcp.Cluster('BigQuery', fillcolor=grey2, style='filled') as gbq:
            gbqf = gbq.File('ssd_uploads.ssd_upload_logs')

        drd_util = gcp.Task('DRD Utility')

        gcp.edge_attrs = {color:'blue', fontcolor:'blue', label:'Read By'}
        gcs >> drd_util >> gbqf

    with d.Cluster('Google Adwords') as gmc:
        gaw = gmc.System('Google Adwords', fillcolor='12')


    vgaF >> trx
    vgaLRH >> trx_header

    fu >> gcst % 'gsutil cp'

    drd_util >> gaw % 'Uploads to'



from pydiagrams.Component import *

# Select from one of these Helpers (PUML recommended)
from pydiagrams.helpers.Graphviz import Helper
#from pydiagrams.helpers.PUML     import Helper
#from pydiagrams.helpers.GraphML     import Helper

SystemFill='#b9CDe5'
Helper.theme = r'w:\plantuml\theme-woolworths.iuml'

# Project attributes
connected_retail = {fillcolor:color3}
martech = {fillcolor:color7}

DPI=72
WIDTH=180/DPI

with ComponentContext(Helper) as x:
    x.attrs = {'width':WIDTH}
    with x.Frame('Amazon Web Services') as aws:
        F=aws.Frame
        with F('S3') as aws_s3:
            B=aws_s3.Node
            aws_s3_img = B('Image and Video Storage')

        with F('RDS') as aws_rds:
            D=aws_rds.Database
            aws_rds_inv = D('Inventory')

    with x.Frame('Data As a Service', fillcolor=SystemFill) as daas:
        
        G=daas.Frame # define a group function

        with G('API Gateway') as daas_api_gw:
            I=daas_api_gw.Interface

        with G('Services') as daas_services:
            F=daas_services.Frame
            with F('Personalisation', **connected_retail) as daas_pers:
                daas_pers_api = daas_api_gw.Interface('Personalisation APIs')
                prec = daas_pers.Component('Product Recs')
                offers = daas_pers.Component('Offers')
                daas_pers_api << daas_pers.all_except(shape=Interface)

            with F('Machine Learning', **martech) as daas_ml:
                daas_ml_api = daas_api_gw.Interface('ML APIs')
                cda = daas_ml.Component('Core Analytics')
                denrich = daas_ml.Component('Data Enrichment')
                seg = daas_ml.Component('Advanced Clustering')
                daas_ml_api << daas_ml.all_except(shape=Interface)

            with F('Customer Services', **connected_retail) as daas_cust:
                daas_cust_api = daas_api_gw.Interface('Customer APIs')
                uci = daas_cust.Component('Unified Customer Identity')
                daas_cust_api << daas_cust.all_except(shape=Interface)

            with F('Inventory', **connected_retail) as daas_inv:
                C=daas_inv.Component
                daas_inv_api = daas_api_gw.Interface('Inventory APIs')
                avail = C('Availability Service')
                inv = C('Inventory Service')
                find_in_store = C('Find In Store')
                if False: #Optionally show more detail
                    inv_monitor = C('Inventory Monitor')
                    inv_db = daas_inv.Database('Inventory DB')
                    inv << inv_monitor << inv_db
                inv << aws_rds_inv

                daas_inv_api << daas_inv.all_except(shape=Interface)

            with F('Pricing', **connected_retail) as daas_price:
                C=daas_price.Component
                daas_price_api = daas_api_gw.Interface('Pricing APIs')
                price = C('Price')
                order_calc = C('Order Calculation')
                promotions = C('Promotions')

                daas_price_api << daas_price.all_except(shape=Interface)


            with F('Product', **connected_retail) as daas_product:
                C=daas_product.Component
                daas_product_apis = daas_api_gw.Interface('Product APIs')

                image_service = C('Images')
                product_feed = C('Product Feeds')
                product_service = C('Product Service')

                daas_product_apis << daas_product.all_except(shape=Interface)

                image_service << aws_s3_img

            with F('Orders', **connected_retail) as daas_orders:
                C=daas_orders.Component
                daas_orders_apis = daas_api_gw.Interface('Order APIs')
                orders = C('Orders')
                fulfilment = C('Fulfillment')

                daas_orders_apis << daas_orders.all_except(shape=Interface)

        daas_dal      = daas.Component('Data Abstraction')
        daas_dal.attrs['width'] = len(daas_services.all()) * WIDTH 

        daas_dal.Id = 'daas_dal'
        daas_services.Id = 'daas_services'
        daas_api_gw.Id = 'daas_api_gw'

        daas_raw                = daas.Node('Raw')
        daas_business_layer     = daas.Node('Business Layer')
        daas_redshift           = daas.Database('Redshift')        
        
        daas_bi = daas.Interface('BI')
        daas_athena = daas.Interface('Athena')

        daas_raw ^ daas_business_layer ^ daas_redshift >= daas_bi
        daas_business_layer >= daas_athena

        daas_dal ^ daas_services.all()
        all_services = daas_services.all()
        daas_business_layer >> daas_dal

        # daas_dal ^ daas_services_in


    with x.Frame('Google Cloud Platform') as gcp:
        ga360 = gcp.Component('GA360')
        dv360 = gcp.Component('DV360')
        sa360 = gcp.Component('SA360')
        gcs = gcp.Node('Cloud Storage')
        gbq = gcp.Database('BigQuery')

        gcs ^ gcp.all_except(gcs) # Link gcs to all other Items in gcp

    with x.Frame('Systems of Engagement') as soe:
        with soe.Frame('Customer Engagement') as soe_ce:
            C=soe_ce.Component
            sfmc = C('SFMC')
            digm = C('[Digital Marketing]')
            dync = C('Movable Ink [Dynamic Content]')            
            cself = C('Customer Self Service')
            awsconn = C('Amazon Connect')

        with soe.Frame('POS / Smart Checkout') as soe_pos:
            C=soe_pos.Component
            pos = C('POS')
            sexp = C('Store Experience')

        with soe.Frame('Mobile') as soe_mobile:
            mobile = soe_mobile.Component('Customer Mobile App')

        with soe.Frame('Online Commerce') as soe_web:
            C=soe_web.Component
            ec = C('Headless UX')
            sli = C('SLI [Search]')
            ecp = C('eCommerce Content Personalisation')

        soe_int = soe.Interface('Interface')
        soe_api = soe.Interface('APIs')

        soe_api ^ soe.all_except(soe_int, soe_api) # Link apis to all SOE systems

    with x.Frame('Integration') as integration:
        C=integration.Component
        ait  = C('AIT')
        infm = C('Informatica')
        
        soe_int << integration.all() # link to all items


    with x.Frame('Systems of Record') as sor:

        F=sor.Frame

        cdc = sor.Interface('Change Database Capture (DMS)')
        saas_dal = sor.Interface('SaaS Connectors')
        dbconn = sor.Interface('Database Connectors')

        with F('Financials') as sor_fin:
            C=sor_fin.Component
            crgfin  = C('CRG OraFin')
            djfin   = C('DJ OraFin')

        with F('Merchandise & Supply Chain Systems') as sor_mer:
            C=sor_mer.Component
            djrms   = C('DJ RMS')
            crgrms  = C('CRG RMS')

        with F('Sales Transaction Systems') as sor_sales:
            C=sor_sales.Component
            edw     = C('EDW')
            djpos   = C('DJ POS')
            crgpos  = C('CRG POS')

        with F('Customer Systems', **martech) as sor_customer:
            sor_customer_int = sor_customer.Interface('Connectors')
            C=sor_customer.Component
            sfdc = C('SFDC')
            srvy = C('GetFeedback [Surveys]')
            cmpgn = C('[Campaign  Management]') 
            loyalty = C('[Loyalty]')           
            identity = C('[Identity and AccessManagement]', **connected_retail)

            sor_customer_int << sor_customer.all_except(sor_customer_int)

        with sor.Frame('eCommerce', **connected_retail) as sor_ecomm:
            sor_ecomm_int = sor_ecomm.Interface('Connectors')
            C=sor_ecomm.Component
            isams = C('iSAMS')
            pim = C('[Product Information Management]')
            dam = C('[Digital Asset Management]')
            search = C('SLI Search')
            cms = C('CMSaaS')
            payment = C('Payment Gateway')
            price_promo = C('[Price and Promotions Engine]')
            sor_ecomm_int << sor_ecomm.all_except(sor_ecomm_int)

        saas_dal <<  sor_ecomm_int
        saas_dal << sor_customer_int

        # link cdc to all items in sor_mer, sor_fin, sor_sales
        db_systems = sor_mer.all() + sor_fin.all_except(djfin) + sor_sales.all()
        #cdc << db_systems
        cdc << dbconn
        #dbconn << db_systems

        cdc ^ daas_raw

        dbconn << db_systems

        daas_dal << [dbconn , saas_dal]
        saas_dal << djfin

        sor_int = sor.Interface('SOR Interfaces')

        sor_int ^ integration.all()
        sor_int << dbconn

    daas_raw <= gcs
    gcs >> daas_raw 

    soe_int ^ sfmc 

    soe_api << daas_api_gw.all()

    infm >> gcs




import sys
from collections import OrderedDict
from functools import wraps

# Get access to the running scripts variables
module_globals = sys.modules['__main__'].__dict__

log_enabled=False

def logged(class_name):
    def actual_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if log_enabled:
                print(f'\n<< Calling: {class_name}.{func.__name__} >>')
            return func(*args, **kwargs)
        return wrapper
    return actual_decorator

class BaseItem(object):
    """ Defines a base item, eg: a Component in a Component Diagram """
    def __init__(self,  label, **attrs):
        self.owner = None
        self.label = label
        self.attrs = attrs
        self.id = None

    def __repr__(self):
        return 'BaseItem.label = "{}"'.format(self.label)

    @property
    def Diagram(self):
        if self.owner:
            return self.owner.diagram
        else:
            raise ValueError('%s has no owner' % (self.__repr__()))

    @property
    def Helper(self):
        return self.Diagram.helper

    @property 
    def Collections(self):
        """ Returns the set of collections the item belongs to """
        c = []
        o = self.owner.diagram
        while o.level > 0:
            c.append(o)
            if hasattr(o, 'owner'):
                o = getattr(o.owner, 'diagram', None)
            else:
                o = None                

        return c

class BaseItemCollection(object):
    cls = 'baseItems.BaseItemCollection'

    """ Represents an Object that owns base items, eg: a Diagram """
    @logged(cls)
    def __init__(self, diagram):
        self.collection = OrderedDict()
        self.diagram = diagram

    @logged(cls)
    def add_item(self, i):
        """ Add BaseItem i to the collection """
        i.owner = self
        self.collection[hash(i)] = i
        return i

    def add_new_item(self, itemType, label, **attrs):
        i = itemType(self, label, **attrs)
        return self.add_item(i)

    def items(self):
        return [(k,v) for (k,v) in list(module_globals.items()) if (isinstance(v, BaseItem))]
        
    def values(self):
        for v in list(self.collection.values()):
            yield v

    def values_by_type(self, t):
        for v in self.values():
            if isinstance(v, t):
                yield v

    def set_ids(self):
        for (k,v) in list(self.items()):
            #print k,v.label
            v.id = k

    def get_first_item(self):
        return list(self.collection.values())[0]

    def len(self):
        return len(self.collection.values())        


from pydiagrams.AWS import AWSContext as Context

# Need to use a helper that can understand AWS syntax, eg: PUML_AWS
from pydiagrams.helpers.PUML_AWS import Helper
# from pydiagrams.helpers.Graphviz_AWS import Helper

# These two lines are only needed for debugging during development
# import pydiagrams.baseItems
# pydiagrams.baseItems.log_enabled=False


with Context(Helper) as c:
    # Define some shortcuts
    BUCKET = c.AmazonSimpleStorageServiceS3_Bucket
    DB = c.Genericdatabase

    Zone = c.Package    # Use the Package container for AWS Zones
    Group = c.Frame     # Use Frame for groups

    # Start diagram here
    


from contextlib import contextmanager # contextmanager decorator (for IntegrationSet)
import copy # required to create integration copies for middleware functionality

import pydiagrams.Model as Model

# constants for properties
url = 'url'
direction= 'direction'

class Namespace:
    pass

class IntegrationFragment():
    """ Temporary class used during the creation of a integration when used with operations,
        eg: SimpleItem1 >> SimpleItem2 % 'Label'
        This class stores the result of the intermediate operation: SimpleItem2 % 'label'
    """
    def __init__(self, simpleItem, description):
        self.simpleItem = simpleItem
        self.description = description

##########################################################################
# Item Classes
class Item(Model.Item):
    def __init__(self, label, **attrs):
        super().__init__(label, **attrs)

    def get_attributes_by_type(self, t):
        """ return local attributes that are instances of type t """
        #return [(k,v) for (k,v) in self.__dict__.items() if isinstance(v, t)]
        return [v for v in list(self.__dict__.values()) if isinstance(v, t)]

    @property 
    def Description(self):
        return self.label        

class SimpleItem(Item):
    """ Items that can be owned by Systems """
    @property
    def isEntity(self):
        return isinstance(self, EntityItem)

    @property
    def isModule(self):
        return isinstance(self, ModuleItem)

    # >> 
    # This creates an internal integration from a ModuleItem to an EntityItem
    # can be called on a list of items or a single item
    def __rshift__(self, itemSet):
        def link(item, description=None):
            if isinstance(item, SimpleItem):
                # Determine integration type, Internal or External
                if self.owner == item.owner:
                    # The owner of the items match, therefore Internal Integration
                    assert isinstance(self.owner, SystemItem), 'Item does not belong to a System'
                    newOwner = self.owner

                    # switch around any erroroneous Entity >> Module defintions
                    if self.isEntity and item.isModule:
                        sourceItem = item
                        destItem = self
                    else:
                        sourceItem = self
                        destItem = item

                else:
                    newOwner = self.owner.get_root_owner().current_context_item
                    assert newOwner is not None, 'Can only add integrations from within a IntegrationSet context'
                    sourceItem = self
                    destItem = item

                # Create the IntegrationItem
                integration = IntegrationItem(
                        sourceSystem = self.owner, sourceItem = sourceItem, 
                        destSystem = item.owner, destItem = destItem, 
                        description=description)
                newOwner.add_integration(integration)
                return integration

            else:
                raise ValueError('>> operator only supported for Modules & Entities')

        if type(itemSet) == list:
            for item in itemSet:
                link(item)
            return None
        elif isinstance(itemSet, IntegrationFragment):
            return link(itemSet.simpleItem, itemSet.description)
        else:
            return link(itemSet)

    # %
    # This operator enables an integration to be constructed with a label/description
    # eg: SimpleItem1 >> SimpleItem2 % 'Description'
    # means: create an integration from SimpleItem1 to SimpleItem2 with description 'Description'
    def __mod__(self, description):
        return IntegrationFragment(self, description)

class ModuleItem(SimpleItem):
    pass

class EntityItem(SimpleItem):
    pass        

class ArchItem(Item):
    """ Item with Integrations (eg: System, IntegrationSet) """
    def __init__(self, label, **attrs):
        super().__init__(label, **attrs)
        self.integrations = []

    def add_integration(self, integration):
        integration.owner = self
        self.integrations.append(integration)

    @property
    def Ints(self):
        return self.integrations

    def get_root_owner(self):
        return self.owner

class SystemItem(ArchItem):
    def __init__(self, label, entities, modules, **attrs):
        super().__init__(label, **attrs)
        self.create_attributes(EntityItem, entities)
        self.create_attributes(ModuleItem, modules)

    @property
    def Entities(self):
        """ Return a dict of the entities in this class """
        return self.get_attributes_by_type(EntityItem)

    @property
    def Modules(self):
        """ Return a dict of the entities in this class """
        return self.get_attributes_by_type(ModuleItem)

    def create_attributes(self, itemType, itemList):
        """ Create an attribute for this each item in itemList.
            This is used to create Modules and Entity references
         """
        if not itemList:
            return
        for item in itemList:
            # Create an instance of the item
            if type(item) == str:
                i = itemType(item)
            elif type(item) == tuple:
                i = itemType(item[1])
                i.Id = item[0]     
            
            i.owner = self
            setattr(self, i.Id, i)            
            #print ('Created {itemType}: "{system}".{item}'.format(itemType=itemType, system=self.label, item=i.Id))

class IntegrationSetItem(ArchItem):
    def __init__(self, label, isMiddleware=False, **attrs):
        super().__init__(label, **attrs)
        self.isMiddleware = isMiddleware
        self.mw_system = None # System that is the middleware

    def set_middleware(self, system):
        print('Enabled middlewares')
        self.isMiddleware = True
        self.mw_system = system

    def add_integration(self, integration):
        if not self.isMiddleware:
            super().add_integration(integration)
        else:
            i1 = copy.copy(integration)
            i2 = copy.copy(integration)

            i1.destSystem   = self.mw_system
            i2.sourceSystem = self.mw_system

            # Add the sourceItem to the middleware system
            if not hasattr(self.mw_system, i1.sourceItem.Id):
                new_item = copy.copy(i1.sourceItem)
                new_item.owner = self.mw_system
                setattr(self.mw_system, new_item.Id, new_item)

            for i in [i1, i2]:
                super().add_integration(i)

class IntegrationItem(Item):
    def __init__(self, 
                 sourceSystem, sourceItem, 
                 destSystem, destItem,
                 description = None,
                 url = None,
                 direction = None):
        super().__init__(label = None)                 
        self.sourceSystem = sourceSystem
        self.sourceItem = sourceItem
        self.destSystem = destSystem
        self.destItem = destItem

        self.description = description
        self.direction = direction
        self.url = url

    # & operator
    # This operator enables an integration to have properties appended to it
    # eg: SimpleItem1 >> SimpleItem2 % 'Description' & {url: 'http://example.com/my-integration', direction='right'}
    def __and__(self, attrs):
        assert type(attrs) == dict, "% must be used with a dictionary. eg: % {url: 'http://example.com/my-integration'}"
        self.__dict__.update(attrs)
        return self

    def __repr__(self):
        return 'owner={owner} {sourceSystem.Id}.{sourceItem.Id} >> {destSystem.Id}.{destItem.Id}'.format(**self.__dict__)

    def goVia(self, system):
        """ Makes an integration go via a certain system.
            If an integrations is A >> B, then this makes the integration A >> system >> B
            It returns the pair: [A >> system, system >> B]
        """            
        i0 = IntegrationItem(self.sourceSystem, self.sourceItem,
                             system)
        #TODO


class UserGroupItem(Item):
    pass

##########################################################################

class ArchitectureModel(Model.Model):
    def __init__(self):
        super().__init__(diagram=None)

        self.integrationSets = []        
        self.current_context_item = None
        self.app_user_groups = Model.ManyToMany()

    def System(self, label, entities=None, modules=None, **attrs):
        """ Return the System type """
        new_item = SystemItem(label, entities, modules, **attrs)
        return self.add_item(new_item)

    def Application(self, *args, **kwargs):
        """ Same as System, provided for compatibility with ARK model """
        return self.System(*args, **kwargs)

    def UserGroup(self, label, **kwargs):
        return self.add_item(UserGroupItem(label))

    @contextmanager
    def IntegrationSet(self, label, isMiddleware=False, **attrs):
        intSet = self.add_item(IntegrationSetItem(label, isMiddleware, **attrs))

        if isMiddleware:
            intSet.set_middleware(self.System(label, **attrs))

        self.current_context_item = intSet
        self.integrationSets.append( intSet )
        yield self.current_context_item
        self.current_context_item = None

    @property
    def Systems(self):
        return [v for v in self.values_by_type(SystemItem)]

    @property
    def IntegrationSets(self):
        return [v for v in self.values_by_type(IntegrationSetItem)]

    @property
    def Integrations(self):
        ints = []
        for IS in self.IntegrationSets:
            ints += IS.Ints
        return sorted(ints, key=lambda x:(x.sourceSystem.Id, x.destSystem.Id))

class ArchitectureContext(Model.Context):
    def __init__(self):
        super().__init__(ArchitectureModel())

    def __exit__(self, *args):
        print ('About to exit')
        super().__exit__(*args)
        print ('Done exit')


# View Data Source diagram

# Standard python libraries
from contextlib import contextmanager #required for the cond function

# Local python libraries
import pydiagrams.Diagram as diagram

##########################################################################
# Item Classes
class DatabaseObjectItem(diagram.Item):
    def __init__(self, label, shape, **attrs):
        attrs.update({'shape':shape})
        diagram.Item.__init__(self, label, **attrs)

        # Stores the objects id's that SEND DATA to this instance
        # eg: if this were a view, the sources would be the tables/views th
        self.sources = []
        
    def add_source(self, source_item):
        #print 'add_source "{}" << "{}"'.format(self.Label,source_item.Label)
        assert isinstance(source_item, DatabaseObjectItem), "source_item is not a DatabaseObjectItem"
        self.sources.append(source_item)
        source_item.link(self)  # Create a link from the source_item to this


    # <<  : Set sources for this item
    def __lshift__(self, i2):
        if type(i2) == list:
            for i in i2:
                self.add_source(i)
        elif isinstance(i2, DatabaseObjectItem):
            self.add_source(i2)
        else:
            raise ValueError

# ========================================================
class ViewDiagram(diagram.DiagramBase):
    nodeTypes = {
        'Table'         :   'Table',
        'View'          :   'parallelogram',
        'Package'       :   'invhouse',
        'File'          :   'folder',
        'Integration'   :   'invtrapezium',
        'System'        :   'ellipse',
        'Task'          :   'cds'
    }

    def __init__(self, helper, filename=None, label=None, **attrs):
        diagram.DiagramBase.__init__(self, helper, filename, label, **attrs)
           
def make_method(shape):
    def make_item(self, label, **attrs):
        return self.add_item(DatabaseObjectItem(label, shape, **attrs))
    return make_item

# Dynamically create methods in ViewDiagram that create Item
for (method, shape) in list(ViewDiagram.nodeTypes.items()):
    setattr(ViewDiagram, method, make_method(method))            

# ========================================================
class ViewContext(diagram.Context):
    def __init__(self, helper, filename=None, **attrs):
        self.diagram = ViewDiagram(helper, filename, **attrs)

# ========================================================
class Items(diagram.Items):
    def __init__(self, *items):
        super(Items, self).__init__(*items)


from pydiagrams.Views  import ViewContext

from pydiagrams.helpers.Graphviz import Helper
#from pydiagrams.helpers.PUML import Helper
from pydiagrams.helpers.constants import *

with ViewContext(Helper, label="Basic Example") as d:

    # Create some Nodes
    t = d.Table('Table')
    v = d.View('View')
    i = d.Integration('Integration')

    # Create an 'edge' from Table to View
    v >> t

    # Use "%" to specify a label for the edge
    i >> t % "Writes to"
    


#colours
white = 'white'

# Graphviz colour scheme: paired12
(color1, color2, color3, color4, color5, color6, color7, color8, color9, color10, color11, color12) = \
    ("#a6cee3","#1f78b4", "#b2df8a", "#33a02c", "#fb9a99", "#e31a1c", "#fdbf6f", "#ff7f00", "#cab2d6", "#6a3d9a", "#ffff99", "#b15928" )

# Graphviz colour scheme: grey6
(grey1, grey2, grey3, grey4, grey5, grey6) = \
    ("#f7f7f7", "#d9d9d9", "#bdbdbd", "#969696", "#636363", "#252525" )

#attributes
(color, fillcolor, fontcolor, label, style, width) = \
    "color fillcolor fontcolor label style width".split(' ')





from pydiagrams.helpers.constants import *
from pydiagrams.helpers import helper 


# Map of shapes
helper.shape = {
        'Cloud':'cloud',
        'Database':'database',
        'Folder':'folder',
        'Frame':'frame',
        'Node':'node',
        'Package':'package',
        'Rectangle':'rectangle',    
        'Component':'component',
        'Interface':'interface',
        'Diamond':'interface',

        # Views
        'Table'         :   'rectangle',
        'View'          :   'rectangle',
        'Package'       :   'package',
        'File'          :   'folder',
        'Integration'   :   'interface',
        'System'        :   'cloud',
        'Task'          :   'component',

        #Architecture
        'Entity'        :   'rectangle',
        'Module'        :   'component',
        'EntityGroup'   :   'database'
    }

helper.comment_format = "' {text}\n"


class Helper(helper.helper):
    """ 
    The Helper class stores the methods specific to a diagram language.
    It's passed to the FlowchartBase object to be used when rendering code
    """

    extension   = "puml"
    name = "PUML"

    arrows = {
         'vertical'   : '-->',
         'horizontal' : '->',
         'right'    : '-right->',
         'left'     : '-left->',
         'up'       : '-up->',
         'down'     : '-down->',
         'hline'    : '-',
         'vline'    : '--',
         'vdotted'  : '..>'
        }

    # For plantuml, the theme to include
    theme = 'w:\\plantuml\\theme-blue.iuml'

    @staticmethod
    def node(id, label, **kwargs):
        #eg: interface "[DR3] Customer Review Authorisation" as rms_cra
        n = str(Helper.shape(kwargs.get('shape', 'rectangle')))
        if label or label != '':            
            n += ' "{}"'.format(label)
        n += ' as {} '.format(id)

        if fillcolor in kwargs:
            FILLCOLOR=kwargs[fillcolor]
            if not FILLCOLOR.startswith('#'):
                n += '#'
            n += FILLCOLOR

        if 'note' in kwargs:
            n += '\nnote right of {id}\n{note}\nend note'.format(id=id, note=kwargs['note'])
        return n

    @staticmethod
    def edge(fromId, toId, label=None, **kwargs):
        dir = kwargs.get('dir', None)

        if not dir:
            dir = 'vertical'

        edge = '{} {} {}'.format(fromId, Helper.arrows[dir], toId)

        if label:
            edge += ' : ' + label

        return edge

    @staticmethod
    def startDiagram(*args, **attrs):
        """ Called first during the render to initialise a diagram """

        ret= """@startuml {filename}""".format(filename=args[0])

        if Helper.theme:
            ret += "\n!include {theme}".format(theme=Helper.theme)

        return ret


    @staticmethod
    def endDiagram(*args):
        """ Called last uring the render to finalise a diagram """
        return "@enduml"


    @staticmethod
    def startSubdiagram(id, label, **kwargs):
        shape=Helper.shape(kwargs.get('shape', 'rectangle'))
        #label=kwargs.get('label')
        
        
        t = shape
        if label and label != '':
            t += ' "{}" '.format(label)

        color=str(kwargs.get('fillcolor', ""))
        if color:
            if not color.startswith('#'):
                t += '#' 
            t += str(color)

        #return '{shape} "{label}" {color} {{\n'.format(shape=shape, label=label, color=color)
        return t + ' {\n'

    @staticmethod
    def endSubdiagram(id, label, **kwargs):
        return "\n}} \n".format(id=id)


from pydiagrams.AWS import AWSContext as Context

# Need to use a helper that can understand AWS syntax, eg: PUML_AWS
# from pydiagrams.helpers.PUML_AWS import Helper
from pydiagrams.helpers.Graphviz_AWS import Helper

# These two lines are only needed for debugging during development
# import pydiagrams.baseItems
# pydiagrams.baseItems.log_enabled=False

EMP='Emp'
CONFIG='Config'

with Context(Helper) as c:
    # Define some shortcuts
    BUCKET = c.AmazonSimpleStorageServiceS3_Bucket
    DB = c.Genericdatabase

    Zone = c.Package    # Use the Package container for AWS Zones
    Group = c.Frame     # Use Frame for groups

    # ---------------------------------------------------
    # Start Diagram

    crg_payroll  = DB("Payglobal")
    crg_peopleDB = DB("People DB")    
    
    with c.Database("RMS DB"):
        workplace_config = c.Component( "Workplace Config - RMSDB")

    with c.Cloud("Empower"):
        dj_payroll = DB( "DJ Employees")

    with Group('AWS Cloud / DaaS'):

        dms      = c.AWSDatabaseMigrationService("AWS Database Migration Service")
        aws_sftp = c.AWSTransferforSFTP("AWS Transfer for SFTP")

        with Zone('Landing') as landing:
            aws_landingEmp      = BUCKET(EMP)
            aws_landingConfig   = BUCKET(CONFIG)

        with Zone('Raw') as raw:
            aws_rawEmp      = BUCKET(EMP)
            aws_rawConfig   = BUCKET(CONFIG)

        aws_glue1 = c.AWSGlue("AWS Glue Emp Job")
        aws_glue2 = c.AWSGlue("AWS Glue Config Job")

        with Zone('Staged') as curated:
            aws_curateEmp       = BUCKET(EMP)
            aws_curateConfig    = BUCKET(CONFIG)

        with Group('Lambda') as lambda_group:
            aws_lambda_fbAPI     = c.AWSLambda("AWS Lambda")
            aws_lambda_fbAPI_new = c.AWSLambda_LambdaFunction("New Employee")
            aws_lambda_fbAPI_mod = c.AWSLambda_LambdaFunction("Modify Employee")
            aws_lambda_fbAPI_grp = c.AWSLambda_LambdaFunction("Group Assignment")

    with c.Cloud('FB Workplace'):
        workplace_fb = DB('Employees')

    ###############################################
    # Connections
    # >> : normal connection 'to'
    # <= : left
    # >= : right
    # %  : Label of the connection

    # On prem HR DBs ------------------------------
    crg_payroll >= crg_peopleDB % 'Informatica (Existing)'

    dj_payroll <= aws_sftp % 'FTP'
    aws_sftp >> aws_landingEmp

    # To dms : '<<' means from
    crg_peopleDB >> dms
    workplace_config >> dms

    dms >> landing.all() # DMS links to all items in raw

    # AWS ----------------------------------------

    aws_landingEmp >> aws_rawEmp
    aws_landingConfig >> aws_rawConfig

    # Link raw to curated via glue
    aws_rawEmp    >> aws_glue1 % 'via S3 event' >> aws_curateEmp
    aws_rawConfig >> aws_glue2 % 'via S3 event' >> aws_curateConfig

    # Link all Curated items to lambda
    curated.all() >> aws_lambda_fbAPI         

    # Lambdas
    aws_lambda_fbAPI >> lambda_group.all_except(aws_lambda_fbAPI)

    # Link lambdas to FB
    lambda_group.all_except(aws_lambda_fbAPI) >> workplace_fb    



import pydiagrams.baseItems as bi

class Item(bi.BaseItem):
    @property 
    def Id(self):
        if self.id:
            return self.id
        else:
            return self.label.replace(' ', '_')

    @Id.setter
    def Id(self, value):
        self.id = value

# Model
class Model(bi.BaseItemCollection):
    pass

class Context():
    """ Represents a general data model. Implements as a context """
    def __init__(self, model):
        self.model = model

    def __enter__(self):
        return self.model

    def __exit__(self, *args):
        self.model.set_ids()

# A class for mantaining many-many relationships between two objects
# the classes are A and B
class ManyToMany(dict):
    class Index(dict):
        def __setitem__(self, key, value):
            if key in self:
                self[key].append(value)
            else:
                super().__setitem__(key, [value])

    def __init__(self):
        super().__init__({})
        self._A = ManyToMany.Index()
        self._B = ManyToMany.Index()

    def get(self, a, b):
        return self.__getitem__((a,b))

    def set(self, a, b, value=None):
        super().__setitem__( (a,b), value)
        self._A[a] = b
        self._B[b] = a

    def get_by_A(self, a):
        """ Returns the set of B objects for an A """
        return self._A.__getitem__(a)

    def get_by_B(self, b):
        """ Returns the set of A objects for a B """
        return self._B.__getitem__(b)

    def set_by_A(self, a, b_list):
        for b in b_list:
            self.set(a, b)

    def set_by_B(self, b, a_list):
        for a in a_list:
            self.set(a, b)            


# This version demonstrates the functionality for Component groups to be 'anonymous'
# ie: instead of
## with ComponentContext(...) as x:
##    with x.Frame('My Group') as MYGROUP:
##       foo = MYGROUP.Component('foo')
# Have this instead:
## with ComponentContext(...) as x:
##    with x.Frame('My Group'):
##          foo = x.Component('foo')

from pydiagrams.Component import *

# Select from one of these Helpers (PUML recommended)
# from pydiagrams.helpers.Graphviz import Helper
from pydiagrams.helpers.PUML     import Helper
#from pydiagrams.helpers.GraphML     import Helper
# from pydiagrams.helpers.arcentry     import Helper


SystemFill='#b9CDe5'
Helper.theme = r'w:\plantuml\theme-woolworths.iuml'

# Project attributes
connected_retail = {fillcolor:color3}
martech = {fillcolor:color7}

DPI=72
WIDTH=180/DPI
PAD=15/DPI

with ComponentContext(Helper) as x:
    x.attrs = {'width':WIDTH}

    # Define some shortcuts for functions
    F=x.Frame
    C=x.Component
    I=x.Interface
    N=x.Node
    D=x.Database

    with F('Amazon Web Services') as aws:
        with F('S3') as aws_s3:
            aws_s3_img = N('Image and Video Storage')

        with F('RDS') as aws_rds:
            aws_rds_inv = D('Inventory')

    with F('Data As a Service', fillcolor=SystemFill) as daas:
        
        with F('API Gateway') as daas_api_gw:
            daas_pers_api       = I('Personalisation APIs')
            daas_ml_api         = I('ML APIs')
            daas_cust_api       = I('Customer APIs')
            daas_inv_api        = I('Inventory APIs')
            daas_price_api      = I('Pricing APIs')
            daas_product_apis   = I('Product APIs')
            daas_orders_apis    = I('Order APIs')

        with F('Services') as daas_services:
            with F('Personalisation', **connected_retail) as daas_pers:
                prec = C('Product Recs')
                offers = C('Offers')
                daas_pers_api << daas_pers.all_except(shape=Interface)

            with F('Machine Learning', **martech) as daas_ml:
                cda =       C('Core Analytics')
                denrich =   C('Data Enrichment')
                seg =       C('Advanced Clustering')
                daas_ml_api << daas_ml.all_except(shape=Interface)

            with F('Customer Services', **connected_retail) as daas_cust:
                uci = C('Unified Customer Identity')
                daas_cust_api << daas_cust.all_except(shape=Interface)

            with F('Inventory', **connected_retail) as daas_inv:
                avail =         C('Availability Service')
                inv =           C('Inventory Service')
                find_in_store = C('Find In Store')
                if False: #Optionally show more detail
                    inv_monitor = C('Inventory Monitor')
                    inv_db = D('Inventory DB')
                    inv << inv_monitor << inv_db
                inv << aws_rds_inv

                daas_inv_api << daas_inv.all_except(shape=Interface)

            with F('Pricing', **connected_retail) as daas_price:
                price =         C('Price')
                order_calc =    C('Order Calculation')
                promotions =    C('Promotions')

                daas_price_api << daas_price.all_except(shape=Interface)

            with F('Product', **connected_retail) as daas_product:
                image_service =     C('Images')
                product_feed =      C('Product Feeds')
                product_service =   C('Product Service')

                daas_product_apis << daas_product.all_except(shape=Interface)
                image_service << aws_s3_img

            with F('Orders', **connected_retail) as daas_orders:
                orders =        C('Orders')
                fulfilment =    C('Fulfillment')

                daas_orders_apis << daas_orders.all_except(shape=Interface)

        daas_dal      = C('Data Abstraction', width=len(daas_services.all()) * (PAD+WIDTH))

        daas_raw                = N('Raw')
        daas_business_layer     = N('Business Layer')
        daas_redshift           = D('Redshift')        
        
        daas_bi = I('BI')
        daas_athena = I('Athena')

        daas_raw ^ daas_business_layer ^ daas_redshift >= daas_bi
        daas_business_layer >= daas_athena

        daas_dal ^ daas_services.all()
        all_services = daas_services.all()
        daas_business_layer >> daas_dal

    with F('Google Cloud Platform') as gcp:
        ga360 = C('GA360')
        dv360 = C('DV360')
        sa360 = C('SA360')
        gcs =   N('Cloud Storage')
        gbq =   D('BigQuery')

        gcs ^ gcp.all_except(gcs) # Link gcs to all other Items in gcp

    with F('Systems of Engagement') as soe:
        soe_int = I('Interface')
        soe_api = I('APIs')

        with F('Customer Engagement') as soe_ce:
            sfmc = C('SFMC')
            digm = C('[Digital Marketing]')
            dync = C('Movable Ink [Dynamic Content]')            
            cself = C('Customer Self Service')
            awsconn = C('Amazon Connect')

        with F('POS / Smart Checkout') as soe_pos:
            pos     = C('POS')
            sexp    = C('Store Experience')

        with F('Mobile') as soe_mobile:
            mobile = C('Customer Mobile App')

        with F('Online Commerce') as soe_web:
            ec  = C('Headless UX')
            sli = C('SLI [Search]')
            ecp = C('eCommerce Content Personalisation')

        # Link apis to all SOE systems
        soe_api ^ soe.all_except(soe_int, soe_api) 

    with F('Integration') as integration:
        ait  = C('AIT')
        infm = C('Informatica')
        
        soe_int << integration.all() # link to all items

    with F('Systems of Record') as sor:
        cdc =       I('Change Database Capture (DMS)')
        saas_dal =  I('SaaS Connectors')
        dbconn =    I('Database Connectors')
        sor_int =   I('SOR Interfaces')

        with F('Financials') as sor_fin:
            crgfin  = C('CRG OraFin')
            djfin   = C('DJ OraFin')

        with F('Merchandise & Supply Chain Systems') as sor_mer:
            djrms   = C('DJ RMS')
            crgrms  = C('CRG RMS')

        with F('Sales Transaction Systems') as sor_sales:
            edw     = C('EDW')
            djpos   = C('DJ POS')
            crgpos  = C('CRG POS')

        with F('Customer Systems', **martech) as sor_customer:
            sor_customer_int = I('Connectors')
            sfdc =      C('SFDC')
            srvy =      C('GetFeedback [Surveys]')
            cmpgn =     C('[Campaign  Management]') 
            loyalty =   C('[Loyalty]')           
            identity =  C('[Identity and AccessManagement]', **connected_retail)

            sor_customer_int << sor_customer.all_except(sor_customer_int)

        with F('eCommerce', **connected_retail) as sor_ecomm:
            sor_ecomm_int = I('Connectors')
            isams =         C('iSAMS')
            pim =           C('[Product Information Management]')
            dam =           C('[Digital Asset Management]')
            search =        C('SLI Search')
            cms =           C('CMSaaS')
            payment =       C('Payment Gateway')
            price_promo =   C('[Price and Promotions Engine]')
            sor_ecomm_int << sor_ecomm.all_except(sor_ecomm_int)

        saas_dal <<  sor_ecomm_int
        saas_dal << sor_customer_int

        # link cdc to all items in sor_mer, sor_fin, sor_sales
        db_systems = sor_mer.all() + sor_fin.all_except(djfin) + sor_sales.all()
        cdc << dbconn
        cdc ^ daas_raw

        dbconn << db_systems

        daas_dal << [dbconn , saas_dal]
        saas_dal << djfin

        sor_int ^ integration.all()
        sor_int << dbconn

    daas_raw <= gcs
    gcs >> daas_raw 

    soe_int ^ sfmc 

    soe_api << daas_api_gw.all()

    infm >> gcs

print ('-'*80)
print(f'showing x: {x=} {type(x)=}')
print(f'{x.parent=}')

print(x.__str__)


from pydiagrams.Component import *
from pydiagrams.helpers.PUML     import Helper


with ComponentContext(Helper) as x:
    F = x.Frame
    P = x.Package
    R = x.Rectangle

    G = lambda l: F(l, fillcolor=color1)
    Sg = lambda l: F(l, fillcolor=color3)
    App = lambda s,l: R('**{s}**\\n{l}'.format(s=s, l=l), fillcolor=white)

    C = lambda l: x.Component(l, fillcolor=white)
    P = lambda s: F(s, fillcolor=color3)

    # Stores and Operations
    with G('Stores and Operations'):
        with Sg('POS'):
            vbs = App('VBS', 'Vision Bean Store')
            vod = App('VOD', 'Vision On Demand')

        fsis = App('FSIS', 'Frontline Sales Incentive Scheme')
        with P('SIM') as sim:
            C('Logistics')
            C('Inventory')
            C('Items')
            C('Stocktake')
            C('In Store Ordering')
            C('Customer Orders')

    # Merchandise
    with G('Merchandise'):
        with P('RMS') as rms:
            C('Merchandise Hierarchy')
            C('Location Management')
            C('Financial Management')
            C('Invoice Matching')
            C('Landed Costs')
            C('Items')
            C('Replenishment')
            C('Pricing')
            C('Suppliers')
            C('Consignments')
            C('Concession Stock')
            C('Ordering')

    # Food
    with G('Food'):
        amos = App('AMOS', 'Food Master Data')
        intactix = App('Intactix', 'Space Planning')
        episys = App('Episys', 'Ticket Printing')
        atria= App('Atria', 'Scales')
    
    


from pydiagrams.helpers.constants import *
#from pydiagrams.helpers.PUML import Helper
import pydiagrams.helpers.PUML as PUML

class Helper(PUML.Helper):
    """ 
    This Helper extends the PUML helper form using the AWS_PlantUML library
    """

    # extension   = "puml"
    name = "PUML_AWS"

    # arrows = {
    #      'vertical'   : '-->',
    #      'horizontal' : '->',
    #      'right'    : '-right->',
    #      'left'     : '-left->',
    #      'up'       : '-up->',
    #      'down'     : '-down->',
    #      'hline'    : '-',
    #      'vline'    : '--',
    #      'vdotted'  : '..>'
    #     }

    # # For plantuml, the theme to include
    theme = ''

    includes = set()

    @staticmethod
    def node(id, label, **kwargs):
        #eg: interface "[DR3] Customer Review Authorisation" as rms_cra
        shape = Helper.shape(kwargs.get('shape', 'rectangle'))

        # if the shape is text, its a normal component, otherwise it's an AWSDiagramItem
        if type(shape) == str:
            return PUML.Helper.node(id, label, **kwargs)

        n = ''
        if shape.name not in Helper.includes:
            Helper.includes.add(shape.name)
            n = f'!includesub Item__{shape.name}\n'

        # print(f'{shape.name=} {type(shape)=} {label=}')

        n += f'{shape.name}({id},"{label}")' 

        if 'note' in kwargs:
            n += '\nnote right of {id}\n{note}\nend note'.format(id=id, note=kwargs['note'])
        return n

    @staticmethod
    def startDiagram(*args, **attrs):
        """ Called first during the render to initialise a diagram """

        ret= """@startuml {filename}
!$AWS_PLANTUML=%getenv("PLANTUML_AWS")
!include $AWS_PLANTUML\common.puml
""".format(filename=args[0])

        if Helper.theme:
            ret += "\n!include {theme}".format(theme=Helper.theme)

        return ret


from pydiagrams.Views  import ViewContext

from pydiagrams.helpers.Graphviz import Helper
#from pydiagrams.helpers.PUML import Helper
from pydiagrams.helpers.constants import *

with ViewContext(Helper, label="Sources") as d:

    # Create some Nodes
    d.node_attrs = {fillcolor:color4} # Specify a color for all Tables
    t1 = d.Table('Table 1')
    t2 = d.Table('Table 2')
    t4 = d.Table('Table 4')
    t3 = d.Table('Table 3')
    t5 = d.Table('Table 5')
 
    d.node_attrs = {fillcolor:color3} # Specify a color for all Views
    v1 = d.View('View 1')
    v2 = d.View('View 2')
    v3 = d.View('View 3')

    # A << B indicates that item A is sourced from B
    # it's useful for items like Views that get data from numerous Tables
    v1 << t1

    # Can specify that a view comes from multiple Tables by using []
    # v2 gets data from t2, t3, t4
    v2 << [t2, t3, t4]

    v3 << [t3, t4, t5]

    # "View 4" is created 'anonymously'. 
    # This is useful for when you don't need to refer to the Node again...
    d.View('View 4', fillcolor=color1) << [v1, v2, v3]

    # ... but if you must refer to it again, you can via it's label
    # As "View 4" was created under item 'd', we can do d[label] to find it
    d.Task('Integration Task', fillcolor=grey1) << d['View 4']



import pydiagrams.baseItems as baseItems
from pydiagrams.helpers.constants import *

import datetime
import sys
import os.path
from contextlib import contextmanager #required for the DiagramBase.subgraph

from pydiagrams.baseItems import logged

class Item(baseItems.BaseItem):
    cls = "Diagram.Item"
    def __init__(self, label, **attrs):
        baseItems.BaseItem.__init__(self, label, **attrs)
        self._links = []
        self.is_group = False

    @property
    def Id(self):
        return self.id or 'n' + str(hash(self))

    @Id.setter
    def Id(self, value):
        self.id = value        

    @property
    def Label(self):
        return self.label or self.Id.replace('_', ' ').title()

    @Label.setter
    def Label(self, value):
        self.label = value        

    @property
    def Note(self):
        return self.attrs.get('note')

    @Note.setter
    def Note(self, value):
        self.attrs['note'] = value

    @property
    def Links(self):
        """ Return the set of items this item is linked to """
        return self._links

    @logged(cls)
    def __str__(self):
        # Call the helper's node render function
        return self.Helper.node(self.Id, self.Label, **self.attrs)

    def link(self, i2, label=None, **attrs):
        #print ('link: {} -> {}'.format(self.Id, i2.Id))
        """ Create a link between this Item and i2 """
        edge_attrs = attrs
        fromItem = self

        #if isinstance(self, DiagramBase):
        if self.is_group:
            edge_attrs.update(ltail='cluster_'+self.id )
            fromItem = self.get_first_item()

        # If the node is required to go to a subgraph, then
        # need to direct to the first item and set the special Graphviz attributes
        # TODO: Use the helper instead        
        if i2.is_group:
            i2 = i2.get_first_item()
            edge_attrs.update(lhead='cluster_' + i2.Diagram.id) # set the target for the cluster

        d = self.Diagram
        assert d,"Uhoh no diagram for {}".format(self.__repr__())

        #print(f'{d=} ')
        #print(f'{d.level=}')
        #print(f'{i2.Diagram=}')
        while (d != i2.Diagram) and (d.level > 0):
            d = d.parent

        e=Edge(fromItem, i2, label, **attrs)
        d.add_item(e)

        # record the object being linked to
        self._links.append(i2)

        return i2

    def link_edge(self, i2, dir=None):
        def do_link(i2):
            if isinstance(i2, baseItems.BaseItem):
                return self.link(i2, dir=dir)
            elif isinstance(i2, ItemEdge):
                return self.link(i2.item, dir=dir, **i2.attrs)        
            else:
                raise ValueError('type(i2)=' + str(type(i2)))
        if type(i2) == list or isinstance(i2, Items):
            #print(f'Do link {self=} {i2=}')
            #a= list(map(do_link, i2))
            a = [do_link(i) for i in i2]
            return a[0]
        else:
            return do_link(i2)


    def __rshift__(self, i2):
        """ Override the >> operator to mean to add an edge between this item and the i2 """
        return self.link_edge(i2)

    def __mod__(self, s):
        """ Override the % operator to add attributes to an edge. Used with the >> operator
            eg: Add a label to an edge:
                n1 >> n2 % "my label"
            eg: Add a label and color to an edge:
                n1 >> n2 % {label:'my label', color:'blue'}
        """
        if type(s) == str:
            return ItemEdge(item=self, label=s)
        elif type(s) == dict:
            return ItemEdge(item=self, attrs=s) 
        else:
            raise ValueError                                    

class DiagramBase(Item):
    cls = 'Diagram.DiagramBase'
    @logged(cls)
    def __init__(self, helper, filename=None, label=None, **attrs):
        Item.__init__(self, label=label, **attrs)
        self.id = 'g'
        self.collection = baseItems.BaseItemCollection(self)

        self.parent = None
        self.helper = helper

        # The full filename of the user's script
        self._user_module_file = os.path.realpath(sys.modules['__main__'].__file__ )

        self._filename = filename
        self._pathname = None

        self.level = 0

        self.edge_attrs = {}
        self.node_attrs = {}

        self.passthrough_index = 0
        self.is_group = True

    # Property functions to deal with the file naming
    @property 
    def FileName(self):
        if self._filename:
            fn = self._filename 
        else:
            fn = os.path.basename(self._user_module_file).replace('.py', '') 
        return fn + '.' + self.helper.extension

    @FileName.setter
    def FileName(self, value):
        self._filename = value

    @property
    def PathName(self):
        if self._pathname:
            return self._pathname
        else:
            return os.path.dirname(self._user_module_file)

    @PathName.setter
    def PathName(self, value):
        self._pathname = value

    # Full file name (path + file + extension) of the file to be written
    @property 
    def OutputFileName(self):
        return os.path.join(self.PathName, self.FileName)

    def add_item(self, item):
        # Determine base attributes of this item
        if isinstance(item, Edge):
            base_attrs = item.fromItem.Diagram.edge_attrs.copy()
            # Remove the label key
            #if base_attrs.has_key(label):
            #    base_attrs.pop(label)
        elif isinstance(item, Item):
            base_attrs = self.node_attrs.copy()
        else:
            base_attrs = {}


        ia = item.attrs.copy()
        item.attrs = base_attrs
        item.attrs.update(ia)

        if 'label' in item.__dict__:
            if not item.label:
                item.label = base_attrs.get(label)

        return self.collection.add_item(item)

    @logged(cls)
    def __str__(self):
        if self.parent:
            return \
                self.helper.startSubdiagram(self.id, self.label, **self.attrs) + '\n' \
                + '\n'.join([ str(v) for v in list(self.collection.values()) ]) \
                + self.helper.endSubdiagram(self.id, self.label, **self.attrs) + '\n' 
        else:
            return 'ViewDiagram({id}, "{label}")'.format(**self.__dict__)

    @logged(cls)
    def render(self):
        # This is the main diagram
        yield self.helper.comment('Generated by pydiagrams @ {}').format(datetime.datetime.now())
        yield self.helper.startDiagram(self.FileName, self.label, **self.attrs) + '\n'

        for v in list(self.collection.values()):
            # print( 'Rendering',v.id , 'label=',v.label or '')
            yield str(v) + '\n'

        yield self.helper.endDiagram() + '\n'

    @logged(cls)
    def dump(self):
        if self._filename =='stdout': # Special mode for writing output to stdout
            for l in self.render():
                print(l)
        else:
            output_filename = self.OutputFileName
            print('Creating: {}'.format(output_filename))
            with open(output_filename, 'w') as f:
                f.writelines(self.render())
                
    def __getitem__(self, label):
        """ Returns an Item with a label """
        for v in list(self.collection.values()):
            if v.label == label:
                return v
        raise ValueError # notfound

    def get_first_item(self):
        for i in list(self.collection.values()):
            if not isinstance(i, DiagramBase):
                return i

    def passthrough(self, line):
        """ Adds a line of native code to the output """
        return self.add_item(PassthroughItem(line))

    @contextmanager
    def Cluster(self, label, **attrs):
        #attrs.update({'label':label})
        sub = self.__class__(helper=self.helper, label=label, **attrs)

        # Set the attributes for the sub diagram
        sub.parent = self
        sub.level = self.level + 1
        sub.node_attrs = self.node_attrs.copy()
        sub.edge_attrs = self.edge_attrs.copy()
        sub.id = hash(self)

        self.add_item(sub)

        yield sub

        # on exit, set the ids
        sub.collection.set_ids()

    # Return all the items in the diagram. recurses into subdiagrams
    def all(self, name_contains=None):
        items = []
        self.collection.set_ids()
        for v in self.collection.values():
            if isinstance(v, DiagramBase) and v.is_group:
                for v2 in v.all(name_contains):
                    items.append(v2)
            elif isinstance(v, Item):
                if name_contains:
                    if name_contains in v.Id:
                        items.append(v)    
                else:
                    items.append(v)
        return Items(*items)

class Edge(baseItems.BaseItem):
    def __init__(self, fromItem, toItem, label, **attrs):
        baseItems.BaseItem.__init__(self, label, **attrs)
        (self.fromItem, self.toItem) = (fromItem, toItem)

    def __str__(self):
        a=self.attrs.copy()
        if label in a:
            a.pop(label)

        # Call the helper's edge render function
        return self.Helper.edge(self.fromItem.Id, self.toItem.Id, label=self.label, **a)

class PassthroughItem(baseItems.BaseItem):
    def __init__(self, line):
        baseItems.BaseItem.__init__(self, label=None)
        self.line = line

    def __str__(self):
        # Call the helper's node render function
        return self.line


class ItemEdge(object):
    """ Class that is returned as a result of: 
        - Item() % str
        - Item() % dict
    """
    def __init__(self, item, label="", attrs={}):
        self.item = item
        self.attrs = attrs
        if label != "":
            self.attrs.update(label=label)

class Items(object):
    """ Container class for Item. Enables operations on a set of Items
    """
    def __init__(self, *items):
        self.itemList = []
        for i in items:
            assert isinstance(i, Item), '<{}> is not an instance of a Item'.format(i)
            self.itemList.append(i)


    def __getitem__(self, index):
        return self.itemList[index]

    def __iter__(self):
        return iter(self.itemList) 

    # Allow a set of Items to be linked to another item
    def __rshift__(self, i2):
        if isinstance(i2, Item) or isinstance(i2, ItemEdge):
            for i in self.itemList:
                i.link_edge(i2)
        else:
            raise ValueError('type of i2 is {}'.format(type(i2)))

class Stack:
    def __init__(self):
        self._stack = []

    def __iter__(self):
        return iter(self._stack)

    @logged('Diagram.Stack')
    def push(self, e):
        # print(f'\tIn push: {self._stack=}')
        self._stack.append(e)

    def pop(self):
        return self._stack.pop()

    def count(self):
        return len(self._stack)

    def current(self):
        return self._stack[-1]

class Context(object):
    """ Class to manage the context of a Diagram.
        Allows the use of the 'with' syntax to automatically Start and End the flowchart """
    def __init__(self, diagram=None):
        self._groups = Stack()
        self.diagram = diagram



    def __enter__(self):
        return self.diagram

    def __exit__(self, *args):
        #print 'Setting Ids'
        self.diagram.collection.set_ids()
        #print 'About to call render()'
        self.diagram.render()
        self.diagram.dump()         



from pydiagrams.Architecture import ArchitectureContext, Namespace

ug = Namespace() #Namespace for user groups

with ArchitectureContext() as a:            


	#================================================================================
	# Systems
	
	# iSAMS -------------------------------------------------------------------------
	es = a.Application('iSAMS'
			,entities = ['Customer Promo', 'Customers', 'Dispatch', 'Dispatch Locations', 'Inventory', 'Orders', 'Prices', 'Product Images', 'Products', 'Promotions', 'Sales', 'Stores', 'Wishlists']
			,modules  = ['Find In Store', 'Fulfilment', 'Gift Cards', 'Gift Registry', 'Order Management', 'Payment Gateway', 'Product Reviews', 'Promotion Management', 'RM Profiler', 'Websites']
			)
	
	# RMS ---------------------------------------------------------------------------
	rms = a.Application('RMS'
			,entities = ['Inventory', 'Inventory Adjustments', 'Orders', 'Prices', 'Product Images', 'Products', 'Sales', 'Shipments', 'Stores', 'Suppliers', 'Transactions', 'Transfers']
			,modules  = ['Data Enrichment', 'Dispatch', 'Find In Store', 'Inventory Management', 'Merchandising', 'OMS', 'SIM']
			)
	# Integrations
	rms.Data_Enrichment >> rms.Products
	rms.Find_In_Store >> rms.Orders
	rms.Inventory_Management >> [rms.Inventory, rms.Inventory_Adjustments, rms.Shipments, rms.Transfers]
	rms.Merchandising >> [rms.Inventory, rms.Prices, rms.Product_Images, rms.Products, rms.Sales, rms.Stores]
	rms.OMS >> [rms.Dispatch, rms.Orders]
	
	# Manhattan WMoS ----------------------------------------------------------------
	wms = a.Application('Manhattan WMoS'
			,entities = ['ASNs', 'Inventory Transactions', 'Items', 'Picks', 'Stores', 'Suppliers', 'Transfers']
			,modules  = ['Despatch', 'Foundation Data', 'Inventory Management', 'Picking', 'Putaway', 'Receiving', 'Wave Processing']
			)
	# Integrations
	wms.Foundation_Data >> [wms.Items, wms.Stores, wms.Suppliers]
	wms.Inventory_Management >> [wms.Inventory_Transactions, wms.Transfers]
	wms.Picking >> wms.Picks
	wms.Receiving >> wms.ASNs
	
	# Salesforce --------------------------------------------------------------------
	sf = a.Application('Salesforce'
			,entities = ['Customer Promo', 'Customers', 'Products', 'Promotions', 'Transactions']
			,modules  = ['Salesforce', 'Salesforce Marketing Cloud']
			)
	# Integrations
	sf.Salesforce >> [sf.Customer_Promo, sf.Customers, sf.Promotions, sf.Transactions]
	sf.Salesforce_Marketing_Cloud >> [sf.Customer_Promo, sf.Customers, sf.Products, sf.Promotions, sf.Transactions]
	
	# EDW ---------------------------------------------------------------------------
	edw = a.Application('EDW'
			,entities = ['Customer Promo', 'Customers', 'Products', 'Promotions', 'Staff', 'Transactions', 'Wishlists']
			)
	# Integrations
	edw.Staff >> edw.Customers
	
	# DPS ---------------------------------------------------------------------------
	dps = a.Application('DPS'
			,entities = ['Payments']
			,modules  = ['Web Services', 'pxPay']
			)
	# Integrations
	dps.Web_Services >> dps.Payments
	dps.pxPay >> dps.Payments
	
	# Payment Acquirers -------------------------------------------------------------
	pay = a.Application('Payment Acquirers'
			,entities = ['AMEX Cards AU', 'AMEX Cards NZ', 'Credit Plans', 'Mastercard AU', 'Mastercard NZ', 'Pay With Points', 'VISA AU', 'VISA NZ']
			,modules  = ['AMEX', 'CBA', 'China Union Pay', 'Paypal']
			)
	# Integrations
	pay.AMEX >> [pay.AMEX_Cards_AU, pay.Credit_Plans, pay.Pay_With_Points]
	pay.CBA >> [pay.Mastercard_AU, pay.VISA_AU]
	
	# Images ------------------------------------------------------------------------
	img = a.Application('Images'
			,entities = ['Product Images', 'Products']
			,modules  = ['Admation', 'Mondo']
			)
	# Integrations
	img.Admation >> img.Mondo
	img.Mondo >> img.Product_Images
	
	# Freight Providers -------------------------------------------------------------
	fr = a.Application('Freight Providers'
			,modules  = ['Aus Post', 'StarTrack']
			)
	
	# Givex -------------------------------------------------------------------------
	gc = a.Application('Givex'
			,entities = ['Accounts', 'Customers', 'Journals']
			,modules  = ['Card Services', 'Corporate Clients']
			)
	# Integrations
	gc.Card_Services >> [gc.Accounts, gc.Customers, gc.Journals]
	gc.Corporate_Clients >> gc.Card_Services
	
	# Gift Card Fulfilment ----------------------------------------------------------
	gcf = a.Application('Gift Card Fulfilment'
			,modules  = ['AB Note', 'Blackhawk']
			)
	
	# Oracle EBS --------------------------------------------------------------------
	ebs = a.Application('Oracle EBS'
			,entities = ['Journals']
			,modules  = ['General Ledger']
			)
	# Integrations
	ebs.General_Ledger >> ebs.Journals
	
	# Suppliers ---------------------------------------------------------------------
	sup = a.Application('Suppliers'
			,entities = ['Inventory', 'Product Images', 'Products', 'Shipments']
			,modules  = ['Drop Ship', 'Supplier']
			)
	# Integrations
	sup.Supplier >> [sup.Drop_Ship, sup.Product_Images, sup.Products, sup.Shipments]
	
	# Point Of Sale -----------------------------------------------------------------
	pos = a.Application('Point Of Sale'
			,entities = ['Customer Promo', 'Customers', 'Gift Registry', 'Orders', 'Promotions', 'Transactions']
			,modules  = ['EFT Client', 'POS', 'VBS', 'VOD']
			)
	# Integrations
	pos.POS >> [pos.EFT_Client, pos.Orders, pos.Transactions]
	pos.VBS >> pos.Promotions
	pos.VOD >> [pos.Customer_Promo, pos.Customers, pos.VBS]
	
	# PC EFTPOS ---------------------------------------------------------------------
	eft = a.Application('PC EFTPOS'
			,entities = ['Bank Tenders', 'Gift Card Tenders']
			,modules  = ['Tender Handler']
			)
	# Integrations
	eft.Tender_Handler >> [eft.Bank_Tenders, eft.Gift_Card_Tenders]
	
	# SPS Commerce ------------------------------------------------------------------
	sps = a.Application('SPS Commerce'
			,modules  = ['Orders', 'Products', 'Shipments']
			)
	
	# Data Enrichment ---------------------------------------------------------------
	de = a.Application('Data Enrichment'
			,modules  = ['BOA', 'COBRA']
			)
	
	# Web ---------------------------------------------------------------------------
	web = a.Application('Web'
			,modules  = [('AV', 'Kleber (Address Validation)')
						, ('Analytics', 'Google Anayltics')
						, ('BV', 'BazaarVoice (Product Reviews)')
						, ('Customer', 'Customer Browser')
						, ('Optimizely', 'Optimizely (A/B Testing)')
						, ('PI', 'SFMC Predictive Intelligence')
						, ('Search', 'SLI (Product Search)')
						, ('Tag', 'Google Web Tag Management')
						]
			)
	# Integrations
	web.Customer >> [web.AV, web.BV, web.Search, web.Tag]
	web.Tag >> [web.Analytics, web.Optimizely, web.PI]
	
	# EFTREC ------------------------------------------------------------------------
	eftrec = a.Application('EFTREC'
			,entities = ['Transactions']
			,modules  = ['Reconciliation']
			)
	# Integrations
	eftrec.Reconciliation >> eftrec.Transactions
	
	#================================================================================
	# Integrations
	
	with a.IntegrationSet('Informatica'):
	# Integrations
		edw.Transactions >> sf.Transactions
		sf.Customer_Promo >> edw.Customer_Promo
		sf.Customers >> edw.Customers
		sf.Promotions >> edw.Promotions
	
	with a.IntegrationSet('Middleware'):
	# Integrations
		dps.Payments >> eftrec.Transactions                     % '#33 CUP Online Transactions'
		edw.Customer_Promo >> es.Customer_Promo                 % '#12 Customer Promo'        & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/100113033/12+Promotion+Customer+EDW+-+iSAMS'}
		edw.Customers >> es.Customers                           % '#5 Customers'              & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/100112338/05+Customers+EDW+-+iSAMS'}
		edw.Transactions >> es.Orders                           % '#34 Save the Sale and GR'  & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/101792031/34+Save+The+Sale+Gift+Registry+Sales+EDW+-+iSAMS'}
		edw.Transactions >> gc.Accounts                         % '#35 Offline Gift Card Purchases' & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/101982207/35+Offline+Gift+Card+Purchases+EDW-+GiveX'}
		edw.Transactions >> gc.Card_Services                    % '#32 Layby/POD Payments'
		es.Customers >> edw.Customers                           % '#29 Customers'             & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/100373326/29+Customers+iSAMS+-+EDW'}
		es.Dispatch >> rms.Dispatch                             % '#13 Dispatch Request'      & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/100113082/13+Order+Dispatch+Request+iSAMS+-+RMS'}
		es.Find_In_Store >> rms.Find_In_Store                   % '#11 Find In Store'         & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/100112958/11+Find+In+Store+iSAMS+-+RMS'}
		es.Orders >> rms.Orders                                 % '#4 Orders'
		es.Sales >> edw.Transactions                            % '#6 Sales'                  & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/100112498/06+Sales+iSAMS+-+EDW'}
		es.Wishlists >> edw.Wishlists                           % '#31 Wishlists'
		gc.Journals >> ebs.General_Ledger                       % '#7 Journals'
		gc.Journals >> eftrec.Transactions                      % '#30 Gift Card Tenders'
		img.Product_Images >> rms.Product_Images                % '#28 Product Image Filenames' & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/100400426/28+Product+Images+Admation-+RMS'}
		pos.Promotions >> es.Promotions                         % '#8 Promotions'             & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/100112464/08+Promotions+VBS+-+iSAMS'}
		rms.Dispatch >> es.Dispatch                             % '#14 Dispatch Response'     & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/100113142/14+Order+Dispatch+Response+RMS+-+iSAMS'}
		rms.Inventory >> es.Inventory                           % '#9 Inventory'              & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/100112609/09+Availability+RMS+-+iSAMS'}
		rms.OMS >> wms.Picks                                    % '#23 Pickticket'            & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/100400109/23+Pick+Tickets+Orders+RMS+-+WMS'}
		rms.Orders >> es.Orders                                 % '#3 Orders'
		rms.Prices >> es.Prices                                 % '#10 Prices'                & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/100112671/10+Prices+RMS+-+iSAMS'}
		rms.Products >> es.Products                             % '#1 Products'               & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/100111916/01+Products+RMS+-+iSAMS'}
		rms.Products >> wms.Items                               % '#18 Item Master'           & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/100373747/18+WMS+Item+Master+Products+RMS+-+WMS'}
		rms.Products >> wms.Items                               % '#19 Item Cross Ref'
		rms.Shipments >> wms.ASNs                               % '#21 ASNs'                  & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/100399001/21+ASNs+Shipments+RMS+-+WMS'}
		rms.Stores >> es.Stores                                 % '#2 Stores'                 & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/100112000/02+Stores+RMS+-+iSAMS'}
		rms.Stores >> wms.Stores                                % '#27 Stores'                & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/100400333/27+Stores+RMS+-+WMS'}
		rms.Suppliers >> wms.Suppliers                          % '#20 Vendor Master'         & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/100395142/20+WMS+Vendor+Master+Suppliers+RMS+-+WMS'}
		rms.Transfers >> wms.Transfers                          % '#22 Store Distro'          & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/100399872/22+Store+Distros+Transfers+RMS+-+WMS'}
		wms.Inventory_Transactions >> rms.Inventory_Adjustments % '#25 PIX_TRAN'              & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/100400006/25+PIX+TRAN+Inventory+Adjustments+WMS+-+RMS'}
		wms.Transfers >> rms.Transfers                          % '#26 Shipment Confs'        & {'url': 'https://davidjones.atlassian.net/wiki/spaces/OR1/pages/100400228/26+Shipment+Confirmations+WMS+-+RMS'}
	
	with a.IntegrationSet('Internal'):
	# Integrations
		de.COBRA >> rms.Data_Enrichment
		edw.Customer_Promo >> pos.Customer_Promo
		edw.Customers >> pos.Customers           % 'Customers'                 & {'direction': 'both'}
		edw.Products >> img.Products
		edw.Products >> sf.Products
		edw.Promotions >> pos.Promotions
		edw.Transactions >> eftrec.Transactions
		edw.Transactions >> rms.Sales
		es.Fulfilment >> fr.Aus_Post             % 'Aus Post Dispatch'
		es.Fulfilment >> fr.StarTrack            % 'StarTrack Dispatch'
		es.Payment_Gateway >> dps.Payments       % 'Transactions'              & {'url': 'x', 'direction': 'both'}
		es.Payment_Gateway >> pay.Paypal
		es.Websites >> web.Customer
		img.Product_Images >> es.Product_Images
		pos.Transactions >> edw.Transactions
		rms.Products >> edw.Products
		sps.Orders >> rms.OMS                    % '#15,16 Orders'             & {'direction': 'both'}
		sps.Products >> rms.Products             % '#17 Products'
		sps.Shipments >> rms.Shipments
		sup.Drop_Ship >> sps.Orders              % 'Orders'                    & {'direction': 'both'}
		sup.Product_Images >> img.Admation
		sup.Products >> de.COBRA
		sup.Products >> sps.Products
		sup.Shipments >> sps.Shipments
	
	with a.IntegrationSet('Third Party'):
	# Integrations
		eft.Gift_Card_Tenders >> gc.Card_Services % 'Gift Card Tenders'         & {'direction': 'both'}
		es.Gift_Cards >> gc.Card_Services         % 'Gift Cards'                & {'direction': 'both'}
		es.Gift_Registry >> pos.Gift_Registry     % 'Gift Registries'
		es.Order_Management >> pos.Orders         % 'Delivery Options'
		es.Products >> web.Search                 % 'Search Integration'
		gc.Card_Services >> gcf.AB_Note           % 'AB Note'
		gc.Card_Services >> gcf.Blackhawk         % 'Blackhawk'
		pos.EFT_Client >> eft.Tender_Handler      % 'Tenders'                   & {'direction': 'both'}
		wms.Despatch >> fr.Aus_Post               % 'Aus Post Dispatch'
		wms.Despatch >> fr.StarTrack              % 'StarTrack Dispatch'
	

# Generate all diagrams            
from pydiagrams.ArchitectureDiagrams import generate_all
import os.path
generate_all(a, os.path.join(os.path.dirname(__file__), 'Architecture-demo-Ecomm'))            
            



from pydiagrams.helpers.constants import *
from pydiagrams.helpers import helper 

# Map of shapes
helper.shape = {

        #Flowchart/Component
        'Rectangle':'box',    
        'Diamond':'flowchart.condition',

        # Component
        'Cloud':'cloud',
        'Database':'flowchart.database',
        'Folder':'roundedbox',
        'Frame':'terminator',
        'Node':'flowchart.database',
        'Package':'flowchart.loopout',
        'Component':'box',
        'Interface':'minidiamond',

        # Views
        'Table'         :   'box',
        'View'          :   'input',
        'Package'       :   'flowchart.loopout',
        'File'          :   'roundedbox',
        'Integration'   :   'flowchart.loopin',
        'System'        :   'cloud',
        'Task'          :   'ellipse',

        #Architecture
        'Entity'        :   'box',
        'Module'        :   'roundedbox',
        'EntityGroup'   :   'database'
    }

helper.comment_format = "// {text}\n"


class Helper(helper.helper):
    """ 
    The Helper class stores the methods specific to a diagram language.
    It's passed to the FlowchartBase object to be used when rendering code
    """

    extension   = "diag"
    name = "Blockdiag"

    # Arrow directions aren't really supported by blockiag
    # http://blockdiag.com/en/blockdiag/examples.html#direction-of-edges
    arrows = {
         'vertical'   : '->',
         'horizontal' : '->',
         'right'    : '->',
         'left'     : '<-',
         'up'       : '->',
         'down'     : '->',
         'hline'    : '--',
         'vline'    : '--',
         'vdotted'  : '->'
        }

    theme = ''

    @staticmethod
    def render_attrs(label, attrs, enclosed=True):
        attrs.update({'label':label or ""})
        return \
            ('[' if enclosed else '') \
            + ', '.join(['{}="{}"'.format(k,v) for (k,v) in list(attrs.items())]) \
            + (']' if enclosed else '') \
            + ';' 

    @staticmethod
    def node(id, label, **kwargs):
        shape = 'shape'
        if shape in kwargs:
            kwargs[shape] = Helper.shape(kwargs[shape])

        if fontcolor in kwargs:
            kwargs['textcolor'] = kwargs.pop(fontcolor)

        if fillcolor in kwargs:
            kwargs['color'] = kwargs.pop(fillcolor)


        if 'note' in kwargs:
            #Not supported
            kwargs.pop('note')

        n = id + ' ' + Helper.render_attrs(label, kwargs)
        return n

    @staticmethod
    def edge(fromId, toId, label=None, **kwargs):
        if 'dir' in kwargs:
            kwargs.pop('dir')
        return '{} -> {}'.format(fromId, toId) + ' ' + Helper.render_attrs(label, kwargs)

    @staticmethod
    def startDiagram(*args, **attrs):
        return "blockdiag {\norientation=portrait;"


    @staticmethod
    def endDiagram(*args):
        """ Called last uring the render to finalise a diagram """
        return "}"


    @staticmethod
    def startSubdiagram(id, label, **kwargs):
        kwargs.setdefault('label',label)
        if 'shape' in kwargs:
            kwargs.pop('shape')
        if fillcolor in kwargs:
            kwargs['color'] = kwargs.pop(fillcolor)

        attrs = "; ".join(['{}="{}"'.format(k,v) for (k,v) in list(kwargs.items())]) +";"
        return """
group {id} {{
{attrs}""".format(id=id, attrs=attrs)

    @staticmethod
    def endSubdiagram(id, label, **kwargs):
        return "\n}} \n".format(id=id)


#! python3

from pydiagrams.Component import ComponentContext, _groups

# Select from one of these Helpers (PUML recommended)
#from pydiagrams.helpers.Graphviz import Helper
#from pydiagrams.helpers.Blockdiag     import Helper
from pydiagrams.helpers.arcentry     import Helper

# Setup a Component diagram called c
# Demonstrates use of Components, Interface and Notes
with ComponentContext(Helper, 'Component-demo-basic') as c:
    fc = c.Component('First Component', note='A note can also\nbe on several lines')
    c.Interface('Data Access') - fc
    fc >>= c.Interface('HTTP', note='Web Service only') % 'Use'

""" 
@startuml

package "Some Group" {
  HTTP - [First Component]
  [Another Component]
}
 
node "Other Groups" {
  FTP - [Second Component]
  [First Component] --> FTP
} 

cloud {
  [Example 1]
}


database "MySql" {
  folder "This is my folder" {
	[Folder 3]
  }
  frame "Foo" {
	[Frame 4]
  }
}


[Another Component] --> [Example 1]
[Example 1] --> [Folder 3]
[Folder 3] --> [Frame 4]

@enduml
"""

with ComponentContext(Helper, 'Component-demo-grouping') as g:
    with g.Package("Some Group") as g1:
        g.Interface('HTTP') - g.Component('First Component')
        ac = g.Component('Another Component')

    with g.Node('Other Groups') as g2:
        g.Interface('FTP') - g.Component('Second Component')
        # Access components by looking for their label
        # rather than accessing via a  variable
        g1['First Component'] >> g2['FTP']

    with g.Cloud() as g3:
        ex1 = g.Component('Example 1')

    with g.Database('MySql') as db:
        with g.Folder('This is my folder') as fg1:
            f3 = g.Component('Folder 3')
        with g.Frame('Foo') as fg2:
            f4 = g.Component('Frame 4')

    ac >> g3['Example 1']
    ex1 >> f3 >> f4 #Can chain links together

# ---------------------------------------------------------
# It is also possible to change arrow direction by adding 
# left, right, up or down keywords inside the arrow
""" 
@startuml
[Component] -left-> left 
[Component] -right-> right 
[Component] -up-> up
[Component] -down-> down
@enduml
"""

with ComponentContext(Helper, 'Component-demo-arrowdir') as a:
    comp = a.Component('Component')
    i=a.Interface # define a shortcut for defining interfaces
    comp <= i('left')
    comp >= i('right')
    comp ^ i('up')
    comp | i('down')


# Test that an item can be either a group or a node

with ComponentContext(Helper, 'Component-demo-groupnode') as b:
    n1 = b.Node('Node')

    # Use a node as a Group
    with b.Node('Group Node') as n2:
        c1 = b.Component('component')
    
    n1 >> c1
    
    with b.Database('RMS') as db:
        rs = b.Rectangle('rag_skus')

    c1 >> rs




import os
import os.path

import pydiagrams.helpers.Graphviz as Graphviz
import pydiagrams.AWS as AWS

class Helper(Graphviz.Helper):
    AWS_ICONS_PATH = os.getenv('AWS_ICONS')

    extension   = 'gv'    
    name = 'Graphviz_AWS'

    @staticmethod
    def node(id, label, **kwargs):
        shape = 'shape'
        if shape in kwargs:
            s=kwargs.pop(shape)

            if isinstance(s, AWS.AWS_Item):
                kwargs['image'] = s.icon.filename.replace(Helper.AWS_ICONS_PATH, '').lstrip(os.path.sep)
            else:
                # Reset properties to defaults for the node
                kwargs[shape] = s
                kwargs['color'] = 'black'
                kwargs['fixedsize'] = 'false'
                kwargs['labelloc'] = 'c'
                kwargs['width'] = ''
                kwargs['height'] = ''
                kwargs['fillcolor'] = '#fefece'
                return Graphviz.Helper.node(id, label, **kwargs)

        #return '\n' + id + ' ' + Helper.render_attrs(label, kwargs)
        return str(Graphviz.GV_Item(id, label=label, **kwargs))

    @staticmethod
    def startDiagram(*args, **attrs):        
        # Add extra attributes to help an AWS diagram look neater
        Graphviz.Helper.graph_defaults.update(
            nodesep=0.8 # Increase spacing between nodes
            ,imagepath = Helper.AWS_ICONS_PATH
        )
        
        Graphviz.Helper.node_defaults.update(
            fillcolor='white',
            color='white',  # Disables the border
            shape='rect',
            fixedsize='true',
            width=1.38,
            height=1.9,     # Lengthen the shape so that label appears underneath
            imagescale='true',
            labelloc='b'    # Label at bottom of shape
        )

        return Graphviz.Helper.startDiagram(*args, **attrs)




######################################################################
# Standard Imports
from pydiagrams.Views  import ViewContext
from pydiagrams.helpers.Graphviz import Helper
from pydiagrams.helpers.constants import *

#####################################################################
# Start Diagram
with ViewContext(Helper, label="Cluster Demo") as d:

    # Create a cluster containing a Table
    with d.Cluster("Cluster 1", fillcolor=color1) as c1:              
        # Create a Table node in c
        t1 = c1.Table('Table #1')                 

    # Create cluster containing a table
    with d.Cluster("Cluster 2", fillcolor=color2) as c2:              
        # Create a Table node in c2
        t2 = c2.Table('Table #2')                 


    # Clusters can be nested
    with d.Cluster('Cluster 3', fillcolor=color3) as c3:
        with c3.Cluster('Cluster 3.1', fillcolor=color4) as c31:  # created with reference to c3
            t31 = c31.Table('Table #3.1')

        t3 = c3.Table('Table #3')

    # Arrow from Table #2 to Table #1
    t1 >> t2 % "t1 >> t2"

    # Arrow from t3 to t31
    t3 >> t31 % "t3 >> t31"

    # Can link clusters
    c2 >> c31 % "c2 >> c31"

    t4 = d.Table("Table #4")

    # Can link clusters to items
    c1 >> t4 % "c1 >> t4"

    # Can link items to clusers
    t4 >> c3 % "t4 >> c3"


import os
import os.path

import glob
import fnmatch
from PIL import Image

ICON_SET_LIGHT = {'path':'PNG_Light', 'file_suffix':'light-bg', 'suffix':True}
ICON_SET_NORMAL = {'path':'PNG Light', 'file_suffix':'light-bg', 'suffix':False}
ICON_SET_DARK  = {'path':'PNG Dark', 'file_suffix':'dark-bg', 'suffix':True}


class AWS_Icon:
    def __init__(self, filename, category):
        self.filename = filename
        self.category = category
        self.icon_set = category.aws_icons.icon_set

    @property 
    def basename(self):
        return os.path.splitext(os.path.basename(self.filename))[0]

    @property
    def dirname(self):
        return os.path.dirname(self.filename)

    @property
    def ext(self):
        """ The extension of the filename """
        return os.path.splitext(self.filename)[-1]

    @property
    def size(self):
        """ Returns the size of the image """
        return  Image.open(self.filename).size

    @property
    def name(self):
        # remove any scale information from the basename
        n = self.basename.split('@')[0] 

        if self.icon_set['suffix']:
            n = n.rstrip(self.icon_set['file_suffix'])
        
        n = n.rstrip('_')

        return n

    @property
    def scale(self):
        """ 
        The scale of the image as reported in the filename. This comes from looking for @[digit]x at the end of the filename
        eg: '@4x.png' would return 4
        """
        parts = self.basename.split('@')

        # if the name doesn't have a '@' then there's no scale, return 1 as the scale
        if len(parts) == 1:
            return 1

        # if the (last) part after the '@' doesn't have an 'x' then we don't have a scale
        if not parts[-1].endswith('x'):
            return 1

        # Attempt to return the scale as a number
        try:
            return int(parts[-1].rstrip('x'))
        except ValueError:
            return 1

    def __repr__(self):
        return self.filename

class AWS_Category:
    def __init__(self, aws_icons, category_path):
        self.aws_icons = aws_icons
        self.category_path = category_path

    def __repr__(self):
        return '{}("{}")'.format(self.__class__.__name__, self.category_path)

    @property
    def name(self):
        return self.category_path

    @property
    def path(self):
        return os.path.join(self.aws_icons.IconBasePath, self.category_path)

    def iter_files(self):
        glob_pattern = os.path.join(self.path, '**/*.png')
        name_filter = self.aws_icons.IconSet_Suffix
        suffix_filter = self.aws_icons.icon_set['suffix']

        I = lambda fn: AWS_Icon(fn, self)

        for fn in glob.iglob(glob_pattern, recursive=True):
            if name_filter in fn:
                if suffix_filter:
                    yield I(fn)
            else:
                if not suffix_filter:
                    yield I(fn)

    def get_files(self):
        return list(self.iter_files())

class NoIconPathError(Exception):
    def __init__(self):
        super().__init__('No Icon Path specified. Set Environment variable AWS_ICONS the path that contains the AWS Icons')

class AWS_Icons:
    def __init__(self, basepath=None, icon_set=ICON_SET_LIGHT):
        """ 
        basepath : The filepath that stores the AWS Icon set
        icon_set : The icons come in two options: 'PNG_Dark', PNG_Light. This is the set to use
        """        
        self.basepath = basepath or os.getenv('AWS_ICONS')
        if not self.basepath:
            raise NoIconPathError()
        self.icon_set = icon_set

    @property
    def IconBasePath(self):
        return os.path.join(self.basepath, self.icon_set['path']) 

    @property 
    def IconSet_Suffix(self):
        return '_' + self.icon_set['file_suffix']

    def iter_categories(self):
        """ The catagories are the subdirectories """
        for dn in os.listdir(self.IconBasePath):
            d = os.path.join(self.IconBasePath, dn)
            if os.path.isdir(d):
                yield AWS_Category(self, dn)

    def get_categories(self):
        return list(self.iter_categories())


if __name__ == '__main__':
    icons = AWS_Icons()

    for c in icons.iter_categories():
        print(c.path)


from pydiagrams.Views  import ViewContext

from pydiagrams.helpers.Graphviz import Helper
#from pydiagrams.helpers.PUML import Helper
from pydiagrams.helpers.constants import *

with ViewContext(Helper, label="Colours and Shapes") as d:
    # setting 'node_attrs' specifies default properties for all new Nodes
    # Graphviz supports fixed widths
    d.node_attrs = {'width':"2.0"}

    # View diagrams support different shapes
    t = d.Table('Table')
    v = d.View('View')
    i = d.Integration('Integration')
    p = d.Package('Package')
    f = d.File('File')
    s = d.System('System')
    k = d.Task('Task')

    # Colours can be set with 'fillcolor' attribute
    # For conveinance, pydiagrams supports the paired12 colour scheme
    # which can be specified by 'colorN' where 1 < N < 12
    tc = d.Table('color1',      fillcolor=color1)
    vc = d.View('color2',       fillcolor=color2)
    ic = d.Integration('color3', fillcolor=color3)
    pc = d.Package('color4', fillcolor=color4)
    fc = d.File('color5', fillcolor=color5)
    sc = d.System('color6', fillcolor=color6)
    kc = d.Task('color7', fillcolor=color7)

    tc2 = d.Table('color8',      fillcolor=color8)
    vc2 = d.View('color9',       fillcolor=color9)
    ic2 = d.Integration('color10', fillcolor=color10)
    pc2 = d.Package('color11', fillcolor=color11)
    fc2 = d.File('color12', fillcolor=color12)

    # pydiagrams also supports the grey6 colorscheme
    # which can be specified by 'greyN' where 1 <= N <- 6
    sg = d.System('grey1', fillcolor=grey1)
    kg = d.Task('grey2', fillcolor=grey2)
    tg = d.Table('grey3',      fillcolor=grey3)
    vg = d.View('grey4',       fillcolor=grey4)
    ig = d.Integration('grey5', fillcolor=grey5)
    pg = d.Package('grey6', fillcolor=grey6, fontcolor=white) #Also set fontcolor


    # setting the global attributes allows new edges to share common properties
    d.edge_attrs = {label:'With fillcolor'}
    # edges can be 'chained together'
    t >> tc >> tc2 >> tg 
    v >> vc >> vc2 >> vg
    i >> ic >> ic2 >> ig
    p >> pc >> pc2 >> pg
    f >> fc >> fc2
    s >> sc >> sg
    k >> kc >> kg

    
    


# Attempting an ideal syntax that uses local variables
# eg:
# instead of defining a component:
#   rms.Component('rms_prod', 'Products')
# do this:
#   rms_prod = rms.Component('Products')

# Lines:
#       >>  link (vertical arrow)
#       >>= dotted link (vertical dotted)
#       -   horizontal line (no arrows)
#       ==  vertical line  (no arrows)
# With Direction:
#       ^   up
#       >=  right
#       <=  left
#       |   down

from pydiagrams.Component import *

with ComponentContextPUML()  as x:

    with x.Database("RMS", fillcolor='adfcad') as rms:
        rms_prod    = rms.Component('Products')
        rms_sales   = rms.Component('Sales')
        rms_pf      = rms.Interface('[DR1] Product Feed')
        rms_sf      = rms.Interface("[DR2] Sales Feed")
        rms_cra     = rms.Interface('[DR3] Customer Review Authorisation')

        rms_prod >> rms_pf
        rms_sales >> rms_sf
        rms_sales >> rms_cra


    with x.Frame("Biztalk 2016") as bt:
        int_pf = bt.Interface('[DI1] Product Feed')


    with x.Cloud("Informatica Cloud") as ic:
        with ic.Frame("Data Integration", fillcolor=white) as di:
            int_sf = di.Interface('[DI2] Sales Feed')
            int_cf = di.Interface('[DI3] Client Feed')

        with ic.Frame("Application Integration", fillcolor=white) as ai:
            inf_cra = ai.Interface("[DA1] Customer Review Authorisation")


    with x.Cloud("BazaarVoice", fillcolor='d5adfc') as bv:
        bv_pie      = bv.Interface("Post-interaction email feed")
        bv_pf       = bv.Interface("Product Feed")
        bv_prod     = bv.Component("Products")
        bv_sales    = bv.Component("Post Interface Events")
        bv_reviews  = bv.Component("Reviews")
        bv_pie  >> bv_sales
        bv_pf   >> bv_prod
        bv_email    = bv.Interface("Outbound Emails")
        bv_scf      = bv.Interface("Standard Client Feed")

        bv_sales >> bv_email
        bv_prod  >> bv_email

        bv_reviews ^ bv_scf #Up arrow

    with x.Cloud("iSAMS", fillcolor='adfcfc') as isams:
        is_prod = isams.Component('Products')
        is_pdp  = isams.Interface('Product Detail Page')
        is_plp  = isams.Interface('Product Listing Page')
        is_pr   = isams.Interface('Product Review Page')

        is_prod ^ is_plp #Up
        is_prod ^ is_pdp #Up


    with x.Cloud("Google Cloud Platform / AWS S3", fillcolor='fcd5ad') as cloud:
        cloud_crm = cloud.Component('Customer Product Reviews')


    with x.Cloud('Salesforce') as sf:
        sf_crm  = sf.Component('Customer Product Reviews', note='Out of Scope')

    # main group
    c = x.Actor('Customer')

    bv_email >> c % "Email"

    rms_pf >> int_pf
    int_pf >> bv_pf % "SFTP"

    rms_sf >> int_sf
    int_sf >> bv_pie % 'SFTP'

    c >> is_pr % "Writes Review"
    is_pr >> bv_reviews

    bv_reviews >> is_plp % "Rating and Review"
    bv_reviews >> is_pdp % "Rating and Review"

    bv_scf ^ int_cf # Up
    int_cf ^ cloud_crm # Up

    sf_crm | cloud_crm %  "via API" #down

    # Customer Review Authorisation
    inf_cra >> rms_cra
    is_pdp >> inf_cra % 'Online Review Auth'





from . import helper

helper.shape = {
        # Components
        'Cloud':'cloud',
        'Database':'database',
        'Folder':'folder',
        'Frame':'tab',
        'Node':'box3d',
        'Package':'package',
        'Rectangle':'rect',    
        'Component':'component',
        'Interface':'circle',
        'Diamond':'diamond',

        # Views
        'Table'         :   'rect',
        'View'          :   'parallelogram',
        'Package'       :   'invhouse',
        'File'          :   'folder',
        'Integration'   :   'invtrapezium',
        'System'        :   'ellipse',
        'Task'          :   'cds',

        #Architecture
        'Entity'        :   'ellipse',
        'Module'        :   'component',
        'SystemIntegration' : 'rect'
    }

helper.comment_format = "# {text}\n"

class Helper(helper.helper):

    extension   = 'graphml'    
    name = 'GraphML'

    @staticmethod
    def render_attrs(label, attrs, enclosed=True):
        attrs.update({'label':label or ""})
        return \
            ('[' if enclosed else '') \
            + '; '.join(['{}="{}"'.format(k,v) for (k,v) in list(attrs.items())]) \
            + (']' if enclosed else '') \
            + ';' 

    @staticmethod
    def node(id, label, **kwargs):
        attrs = kwargs
        print('{}: fillcolor={}'.format(id, attrs.get('fillcolor')))
        for (k,v) in [('id', id), ('label', label),
                      ('shape', 'rectangle'), 
                      ('fillcolor', '#FFFFFF'),
                      ('height', 30), 
                      ('width', 120),
                      ('x', 0),
                      ('y', 0),
                      ('rotationAngle', 0.0),
                      ('fontSize', 12)]:
            attrs.setdefault(k, v)

        return """
	<node id="{id}">
		<data key="d0">
			<y:ShapeNode>
				<y:Geometry height="{height}" width="{width}" x="{x}" y="{y}"/>
				<y:Fill color="{fillcolor}" transparent="false"/>
				<y:BorderStyle color="#000000" type="line" width="1.0"/>
				<y:NodeLabel alignment="center" autoSizePolicy="content" fontFamily="Dialog" fontSize="{fontSize}" fontStyle="plain" hasBackgroundColor="false" hasLineColor="false" modelName="internal" modelPosition="c" rotationAngle="{rotationAngle}" textColor="#000000" visible="true" >{label}</y:NodeLabel>
				<y:Shape type="{shape}"/>
            </y:ShapeNode>
		</data>		
		<data key="d1">{label}</data>
		<data key="d6">{id}</data>
	</node>
""".format(**kwargs)

    @staticmethod    
    def edge(fromId, toId, label=None, **kwargs):
        kwargs['label']=label
        kwargs['EDGE_ID']= fromId + "-" + toId
        kwargs['source'] = fromId
        kwargs['target'] = toId

        kwargs.setdefault('edge_color', '#000000')

        SOURCE="none"
        TARGET="none"
        edge_dir = kwargs.get('direction', 'right') or 'right'

        standard = 'standard'

        if edge_dir == 'left':
            SOURCE = standard
        elif edge_dir == 'right':
            TARGET = standard
        elif edge_dir == 'both':
            (SOURCE, TARGET) = (standard, standard)

        kwargs.setdefault('SOURCE', SOURCE)
        kwargs.setdefault('TARGET', TARGET)

        e = """
        <edge id="{EDGE_ID}" source="{source}" target="{target}">
            <data key="d2">
                <y:PolyLineEdge>
                    <y:LineStyle color="{edge_color}" type="line" width="1.0"/>
                    <y:Arrows source="{SOURCE}" target="{TARGET}"/>
        """.format(**kwargs)

        if label:
            e += """
                        <y:EdgeLabel alignment="center" distance="2.0" fontFamily="Dialog" fontSize="12" fontStyle="plain" hasBackgroundColor="false" hasLineColor="false" modelName="six_pos" modelPosition="tail" preferredPlacement="anywhere" ratio="0.5" textColor="#000000" visible="true">{label}</y:EdgeLabel>
            """.format(**kwargs)

        e += """
                    <y:BendStyle smoothed="false"/>
                </y:PolyLineEdge>
            </data>
        """.format(**kwargs)

        if label:
            e += '<data key="d3">{label}</data>'.format(label=label)
        if 'url' in kwargs:
            e += '<data key="d5"><![CDATA[{url}]]></data>'.format(**kwargs)        

        e += '</edge>'

        return e

    note_node_attrs = {
            'shape':'note',
            'fillcolor':'lightyellow',
            'color':'darkred',
            'fontname':'Cambria',
            'fontcolor':'darkred' 
            }

    note_edge_attrs = {
            'color':'darkred',
            'style':'dashed',
            'constraint':'false',
            'arrowhead':'open'
    }

    comment_format = "' {text}\n"

    @staticmethod
    def startDiagram(*args, **attrs):
        return """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<graphml xmlns="http://graphml.graphdrawing.org/xmlns" xmlns:java="http://www.yworks.com/xml/yfiles-common/1.0/java" xmlns:sys="http://www.yworks.com/xml/yfiles-common/markup/primitives/2.0" xmlns:x="http://www.yworks.com/xml/yfiles-common/markup/2.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:y="http://www.yworks.com/xml/graphml" xmlns:yed="http://www.yworks.com/xml/yed/3" xsi:schemaLocation="http://graphml.graphdrawing.org/xmlns http://www.yworks.com/xml/schema/graphml/1.1/ygraphml.xsd">
	<key for="node" id="d0" yfiles.type="nodegraphics"/>
	<key attr.name="description" attr.type="string" for="node" id="d1"/>
	<key for="edge" id="d2" yfiles.type="edgegraphics"/>
	<key attr.name="description" attr.type="string" for="edge" id="d3"/>
	<key for="graphml" id="d4" yfiles.type="resources"/>  
	<key attr.name="url" attr.type="string" for="edge" id="d5"/>
	<key attr.name="archid" attr.type="string" for="node" id="d6"/>
	<graph id="G" edgedefault="directed">
        """.format(fontname='Calibri', attrs=Helper.render_attrs(args[1], attrs, False))

    @staticmethod
    def endDiagram(*args):
        return """
	</graph>
</graphml>
"""

    @staticmethod
    def startSubdiagram(id, label, **kwargs):
        attrs = kwargs
        attrs['id'] = id
        #TODO: support fillcolor and other attributes
        return """
	<node id="sub_{id}">
		<graph id="{id}"  edgedefault="directed" >
""".format(**attrs)


    @staticmethod
    def endSubdiagram(id, label, **kwargs):
        return """
		</graph>
	</node>
"""



from . import helper

import json

helper.shape = {
        # Components
        'Cloud':'generic.cloud',
        'Database':'generic.database',
        'Folder':'generic.generic-physical-data-transfer-device',
        'Frame':'generic.bucket',
        'Node':'generic.block',
        'Package':'generic.generic-container',
        'Rectangle':'generic.generic-cube',    
        'Component':'generic.processor',
        'Interface':'generic.gateway',
        'Diamond':'diamond',
        'Server':'generic.server',
        'Actor':'generic.pc',

        # Below here are TODO
        # Views
        'Table'         :   'rect',
        'View'          :   'parallelogram',
        'Package'       :   'invhouse',
        'File'          :   'folder',
        'Integration'   :   'invtrapezium',
        'System'        :   'ellipse',
        'Task'          :   'cds',

        #Architecture
        'Entity'        :   'ellipse',
        'Module'        :   'component',
        'SystemIntegration' : 'rect'
    }

helper.comment_format = "# {text}\n"

class Helper(helper.helper):

    extension   = 'arcentry.json'
    name = 'arcentry'

    node_count = 0

    json_encoder = json.JSONEncoder()

    @staticmethod
    def render_attrs(label, attrs, enclosed=True):
        attrs.update({'label':label or ""})
        return Helper.json_encoder.encode(attrs)
        #return attrs

    @staticmethod
    def node(id, label, **kwargs):
        s= 'shape'
        Helper.remove(kwargs, 'width')

        kwargs.setdefault('type', 'component')
        if s in kwargs:
            kwargs.setdefault('componentType', Helper.shape(kwargs.pop(s)))

        if 'groups' in kwargs:
            kwargs['groups'].reverse()

        if 'links' in kwargs:
            links = kwargs.pop('links')
            if len(links) > 0:
                kwargs.setdefault('connections', links)

        ret = '"{}":{}'.format( id, Helper.render_attrs(label, kwargs) )

        if Helper.node_count > 0:
            ret = ',' + ret
        
        Helper.node_count += 1
        return ret


    @staticmethod    
    def edge(fromId, toId, label=None, **kwargs):
        #return '{} -> {}'.format(fromId, toId) + ' ' + Helper.render_attrs(label, kwargs)
        return ''

    comment_format = "' {text}\n"

    @staticmethod
    def startDiagram(*args, **attrs):
        return "{"

    @staticmethod
    def endDiagram(*args):
        return '}'

    @staticmethod
    def startSubdiagram(id, label, **kwargs):
        # arcentry doesn't have subdiagrams, instead the nodes report their parent groups
        return ''


    @staticmethod
    def endSubdiagram(id, label, **kwargs):
        return ''

    @staticmethod
    def comment(text):
        # comments are not supported in json
        return ''    


# Demonstration of using the my python library 
# for doing flowcharts from python
from pydiagrams.Flowchart import FlowchartContext

from pydiagrams.helpers.Blockdiag import Helper
#from pydiagrams.helpers.PUML import Helper

# Start the diagram, define the output file
with FlowchartContext(Helper) as f:
    # Define some Activity Nodes. 
    # These don't reqire labels, they are generated from the Id 
    application     = f.Activity()
    refactor        = f.Activity()
    container       = f.Activity()
    serverless      = f.Activity()
    virtual_machine = f.Activity('Custome label for VM')

    # Define the Condition Nodes, with labels
    if_serverless       = f.Condition('Can it be a Serverless function?')
    if_long_running     = f.Condition( "Is it a long running function")
    if_run_occasionally = f.Condition( "Does it run occasionally?")
    if_time_sensitive   = f.Condition( "Is it time sensitive?")
    if_10m              = f.Condition( ">10 million requests?")
    if_containers       = f.Condition( "Are Containers supported?")

    # Draw the links between Nodes 
    application >> if_serverless
    refactor >> application

    # Simpler syntax for branching
    # Syntax Option 1: 
    if_serverless.Branch(Yes = if_long_running, No =  if_containers)

    # Syntax Option 2 
    if_long_running.Branch(Yes = refactor, No = if_run_occasionally)


    if_run_occasionally.Branch(
        Yes = if_10m,
        No  = if_time_sensitive)

    if_10m.Branch(
        Yes =  if_containers,
        No =   serverless)

    if_time_sensitive.Branch(
        Yes = if_10m,
        No  = serverless)

    if_containers.Branch(
        Yes = container,
        No  = virtual_machine
    )

    # Pass graphviz code to align the serverless, vm and container nodes
    if Helper.name == 'Graphviz':
        f.passthrough('{rank=max; serverless container virtual_machine};')




# Flowchart diagram

# Standard python libraries
from contextlib import contextmanager #required for the cond function

# Local python libraries
#from py import DiagramBase, Context, Item, PassthroughItem
import pydiagrams.Diagram as diagram
from pydiagrams.helpers.constants import *

def label_format(s):
    """ Converts an id into a label """
    return s.replace('_', ' ').title()

##########################################################################
# Item Classes

class FlowchartDiagramItem(diagram.Item):
    """ Base class for all Items in a Flowchart Diagram """
    def __init__(self, label, shape, **attrs):
        attrs.update({'shape':shape})
        diagram.Item.__init__(self, label, **attrs)
        self.nodeBranches = None

    def Branch(self, **branches):
        self.nodeBranches = branches
        for (label, target) in list(branches.items()):
            self.link(target, label=label_format(label))

class ActivityItem(FlowchartDiagramItem):
    def __init__(self, label, **attrs):
        attrs.setdefault(fillcolor, color1)
        FlowchartDiagramItem.__init__(self, label, shape='Rectangle', **attrs)

class ConditionItem(FlowchartDiagramItem):
    def __init__(self, label, **attrs):
        FlowchartDiagramItem.__init__(self, label, shape='Diamond', **attrs)

class FlowchartDiagram(diagram.DiagramBase):
    def __init__(self, helper, filename=None, **attrs):
        diagram.DiagramBase.__init__(self, helper, filename, **attrs)

    def Activity(self, label=None, **attrs):
        """ Returns an Activity Item """
        return self.add_item(ActivityItem(label, **attrs)) 

    def Condition(self, label, **attrs):
        return self.add_item(ConditionItem(label, **attrs))


class FlowchartContext(diagram.Context):
    def __init__(self, helper, filename=None):
        self.diagram = FlowchartDiagram(helper, filename)


from pydiagrams.Architecture import Item
from pydiagrams.helpers.GraphML import Helper
from pydiagrams.helpers import constants as C

import itertools
import colour # from pip install colour. Doco @ https://pypi.org/project/colour/

class Renderer():
    def __init__(self, filename=None):
        self.lines = []
        self.filename= filename

    def write(self, line):
        self.lines.append(line)

    def dump(self):
        if self.filename:
            fn = self.filename + '.' + Helper.extension
            with open(fn, 'w') as f:
                for line in self.lines:
                    f.write(line + '\n')
            print(f'Done writing to: {fn}')

        else:
            print('\n'.join(self.lines))

    def __enter__(self):
        self.write(Helper.startDiagram('', ''))
        return self

    def __exit__(self, *args):
        self.write(Helper.endDiagram())
        self.dump()

class ItemRenderer():
    def __init__(self, mainRenderer, item, **attrs):
        self.mainRenderer = mainRenderer
        self.item = item
        self.lines = []
        self.attrs = attrs

    def __str__(self):
        return str(self.item.id)       

    def write(self, line):
        self.mainRenderer.lines.append(line)

    def __enter__(self):
        #print('{t}.{id}.begin'.format(t=type(self.item), id=self.item.Id))
        return self

    def __exit__(self, *args):
        #print('{t}.{id}.end'.format(t=type(self.item), id=self.item.Id))
        pass

class SystemRenderer(ItemRenderer):
    def __enter__(self):
        self.write(Helper.startSubdiagram(self.item.Id, self.item.label, **self.item.attrs))
        return super().__enter__()

    def __exit__(self, *args):
        self.write(Helper.endSubdiagram(self.item.Id, self.item.label, **self.item.attrs))
        super().__exit__(*args)

class ShapeRenderer(ItemRenderer):
    """ Renders an item as a basic shape """
    def __init__(self, shape, *args, **attrs):
        super().__init__(*args, **attrs)
        self.shape = shape

    def __str__(self):
        node_id =''
        if hasattr(self.item.owner, 'Id'):
            node_id = self.item.owner.Id

        attrs = self.attrs.copy()
        attrs.update(self.item.attrs)

        node_id += self.item.Id if hasattr(self.item, 'Id') else self.item.id 
        return Helper.node(node_id, self.item.label, shape=self.shape, **attrs)


class EntityRenderer(ShapeRenderer):
    def __init__(self, *args):
        super().__init__('Entity', *args, fillcolor=C.color1)

class ModuleRenderer(ShapeRenderer):
    def __init__(self, *args):
        super().__init__('Component', *args, fillcolor=C.color3)

class SystemShapeRenderer(ShapeRenderer):
    def __init__(self, *args, **kwargs):
        super().__init__('SystemIntegration', *args, **kwargs)


class IntegrationRender(ItemRenderer):
    """ Integration between Items """
    def __str__(self):
        e=self.item
        return Helper.edge(
            fromId = e.sourceSystem.Id + '_' + e.sourceItem.Id
            ,toId = e.destSystem.Id + '_' + e.destItem.Id
            ,label=e.description
            ,url=e.url
            ,direction=e.direction
        )

class IntegrationSystemRender(ItemRenderer):
    """ Integration between Systems """
    def __str__(self):
        e=self.item
        return Helper.edge(
            fromId = e.sourceSystem.Id
            ,toId = e.destSystem.Id
            ,label=e.description
            ,url=e.url
            ,direction=e.direction
        )


class ArchitectureDiagram():
    """ General diagram.
    This class will represent the standard properties and methods for a diagram,
    eg:
    - Filename and File handling
    - Standard properties, etc...
    """
    def __init__(self, model, filename=None, **attrs):
        self.model = model
        self.filename = filename
        self.attrs = attrs

    def generate(self):
        pass


class Entity(ArchitectureDiagram):

    def generate(self):
        m = self.model

        with Renderer(self.filename) as r:
            for system in m.Systems:
                with SystemRenderer(r, system):

                    r.write(Helper.startSubdiagram(system.Id + '_group', '', shape='EntityGroup'))                        
                    for entity in system.Entities:
                        r_ent =  EntityRenderer(r, entity)
                        r.lines += [str(r_ent)]

                    r.write(Helper.endSubdiagram(system.Id + '_group', '', shape='EntityGroup'))                        

                    for module in system.Modules:
                        r.write(str(ModuleRenderer(r, module)))

                    for int in system.Ints:
                        r.write(str(IntegrationRender(r, int)))

            intset_id = 0
            for intsets in m.IntegrationSets:

                for int in intsets.Ints:
                    r.write(str(IntegrationRender(r, int)))

                intset_id += 1


class Integration(ArchitectureDiagram):
    def generate(self):
        m=self.model

        with Renderer(self.filename) as r:
            
            # Go through the integrations and count the number of times a system
            # is the source or destination of the integration.
            # This can be used to determine the width/height of the system node
            int_from = {}
            int_to = {}
            for int in m.Integrations:
                int_from[int.sourceSystem]   = int_from.get(int.sourceSystem, 0) + 1
                int_to[int.destSystem]     = int_to.get(int.destSystem, 0) + 1


            for system in m.Systems:
                size=max(system.int_from, system.int_to)
                r.write(str(SystemShapeRenderer(r, system, 
                    width=60, 
                    height=90+(size*30),
                    rotationAngle=270,
                    fontsize=20
                    )))

            for intsets in m.IntegrationSets:
                for int in intsets.Ints:
                    r.write(str(IntegrationSystemRender(r, int)))

def get_fillcolor(item):
    """ Gets the fillcolor for an item """
    assert isinstance(item, Item), 'item must be of instance Architecture.Item'
    return item.attrs['fillcolor']   

def set_fillcolor(item, color):
    """ Gets the fillcolor for an item """
    assert isinstance(item, Item), 'item must be of instance Architecture.Item'
    item.attrs['fillcolor'] = color   

class System(ArchitectureDiagram):
    def generate(self):
        m=self.model

        with Renderer(self.filename) as r:
            int_count = {}
            for int in m.Integrations:
                int_count[int.sourceSystem]   = int_count.get(int.sourceSystem, 0) + 1
                int_count[int.destSystem]     = int_count.get(int.destSystem, 0) + 1


            for (n,system) in enumerate(m.Systems):
                r.write(str(ShapeRenderer('circle', r, system, width=int_count.get(system) / 5 , fillcolor=get_fillcolor(system))))

            for ((sourceSystem, destSystem), g) in itertools.groupby(m.Integrations, key=lambda x:(x.sourceSystem, x.destSystem)):
                label = ',\n'.join([i.description or i.sourceItem.Id for i in g])
                
                line_color = list(get_fillcolor(sourceSystem).range_to(get_fillcolor(destSystem), 3))[1]
                font_color = list(get_fillcolor(sourceSystem).range_to(get_fillcolor(destSystem), 3))[1]

                line_color.luminance = 0.3
                font_color.luminance = 0.25


                r.write(Helper.edge(fromId = sourceSystem.Id, toId=destSystem.Id, label=label, color=line_color, fontcolor=font_color))   

def assign_system_colors(model, saturation=0.9, lightness=0.8):
    sysCount = len(model.Systems)
    print('Assigning system colors:')
    for (n,system) in enumerate(model.Systems):
        set_fillcolor( system, colour.Color(hsl=( (n/sysCount)*1, saturation, lightness )))
        print('\t{}\tSystem {}, \tfillcolor={}'.format(n,system.id, get_fillcolor(system)))

def count_integrations(model):
    for s in model.Systems:
        s.int_from = 0
        s.int_to = 0

    for int in model.Integrations:
        int.sourceSystem.int_from += 1
        int.destSystem.int_to += 1


def generate_all(model, basefilename=None):    
    assign_system_colors(model)
    count_integrations(model)

    for Diagram in ArchitectureDiagram.__subclasses__():
        print('Generating Diagram type: ', Diagram.__name__)
        Diagram(model, basefilename + '-' + Diagram.__name__).generate()


from setuptools import setup, find_packages

# References for setup tools:
#  https://setuptools.readthedocs.io/en/latest/
# Building and Distributing Packages with Setuptools: https://setuptools.readthedocs.io/en/latest/setuptools.html


setup(name='pydiagrams',
      version='0.4',
      description='Generate software diagrams using Python syntax',
      #url='http://github.com/storborg/funniest',
      author='Matthew Billington',
      author_email='mbillington@gmail.com',
      license='MIT',
      packages=find_packages(),
      install_requires = ['colour'],
      # entry points defines the command line version
      entry_points = {},
      zip_safe=False,
      include_package_data=True
      )



# Local python libraries


# import pydiagrams.Diagram as Diagram
import pydiagrams.Diagram as Diagram
import pydiagrams.Component as Component
from pydiagrams.baseItems import logged
import pydiagrams.utils.AWS_Icons as AWS_Icons


# Maps an AWS Icon to a value that we can use here
class AWS_Item:
    def __init__(self, icon):
        self.icon = icon

    @property
    def name(self):
        n=self.icon.name #.title()
        for repchar in '-.':
            n=n.replace(repchar, '')
        n=n.replace(' ', '_')
        
        return n

    def __str__(self):
        return self.name
        
    def __repr__(self):
        return f'{self.__class__.__name__}({self.name=})'

    @staticmethod
    def from_icons(aws_icons):
        for cat in aws_icons.iter_categories():
            for icon in cat.iter_files():
                if icon.name not in ( 'Database', 'Cloud'):
                    yield AWS_Item(icon)


##########################################################################
# Item Classes
class AWSDiagramItem(Component.ComponentDiagramItem):
    def __init__(self, label, aws_item, **attrs):
        # super().__init__(label, 'rectangle', **attrs)
        try:
            print(f'{aws_item.name=}')
            super().__init__(label, aws_item.name, **attrs)
            self.aws_item = aws_item
        except:
            print(f'{aws_item=}')
            raise

##########################################################################
# Diagram Classes

class AWSDiagram(Component.ComponentDiagram):
    cls = 'AWS.AWSDiagram'

    @logged(cls)
    def __init__(self, helper, filename=None, context=None, aws_icons=None, **attrs):
        super().__init__(helper, filename, context, **attrs)

        # Add each AWS Icon as a method to this package
        for i in AWS_Item.from_icons(aws_icons):
            #print(f'{i.name=}, {i.icon.category.name=}')
            setattr(self, i.name, self.__item__(i))

    @logged(cls)
    def add_new_item(self, aws_item, label, **kwargs):
        if isinstance(aws_item, AWSDiagramItem):
            """ Override the base add_new_item method """
            i = AWSDiagramItem(aws_item=aws_item, label=label, **kwargs)
            collection = self._context._groups.current().collection 
            r= collection.add_item(i)
            return r
        else:
            return super().add_new_item(aws_item, label, **kwargs)

    def __item__(self, item):
        """ Called when a AWS Item is invoked """
        @logged(AWSDiagram.cls)
        def add(label="", **attrs):
            # print(f'{AWSDiagram.cls}.__item__: Calling add for: {item} {label=}')
            return self.add_new_item(item, label, **attrs)
        return add


##########################################################################
# Context Classes
class AWSContext(Diagram.Context):
    def __init__(self, helper, filename=None, **attrs):
        super().__init__()

        # AWS_ICON_PATH=r'D:\Software\plantuml\AWS-Architecture-Icons_PNG\AWS-Architecture-Icons_PNG_20191031'
        # aws_icons = AWS_Icons.AWS_Icons(AWS_ICON_PATH)
        aws_icons = AWS_Icons.AWS_Icons()
        self.diagram=AWSDiagram(helper, filename, context=self,aws_icons=aws_icons, **attrs)


if __name__ == '__main__':
    from pydiagrams.helpers.PUML import Helper
    with AWSContext(Helper) as a:
        pass



# Component diagram

# Standard python libraries
from contextlib import contextmanager #required for the cond function

# Local python libraries
#from Diagram import DiagramBase, Item, Context
import pydiagrams.Diagram as Diagram
from pydiagrams.helpers.PUML import Helper

from .helpers.constants import *

##########################################################################
# Constants

# component shape keys
Component = 'Component'
Interface = 'Interface'


# group shapes
Cloud = 'Cloud'
Database = 'Database'
Folder = 'Folder'
Frame = 'Frame'
Node = 'Node'
Package = 'Package'
Rectangle = 'Rectangle'

##########################################################################
# Item Classes

class ComponentDiagramItem(Diagram.Item):
    """ Base class for all Items in a component Diagram """
    def __init__(self, label, shape, **attrs):
        Diagram.Item.__init__(self, label, **attrs)
        self.attrs['shape'] = shape

    def __hash__(self):
        return abs(hash(self.label))

    def __str__(self):
        attrs = self.attrs.copy()
        attrs['groups'] = [g.id for g in self.Collections]
        attrs['links'] = [g.id for g in self.Links]
        # Call the helper's node render function
        return self.Helper.node(self.Id, self.Label, **attrs)

    # Override operators

    # >>  ==  vertical link
    def __rshift__(self, i2):
        """ Override the >> operator to mean to add an edge between this item and the i2 """
        return self.link_edge(i2, 'vertical')

    def __lshift__(self, i2):
        """ Override the << operator to mean to add an edge between i2 this item """
        def link(source, dest):
            return source.link_edge(dest, 'up')

        if type(i2) == list:
            l=[link(i, self) for i in i2]
            return l[0]
        else:
            return link(i2, self)

    # ^  == up link
    def __xor__(self, i2):
        """ Override the ^ operator to mean to add an 'up' edge """
        return self.link_edge(i2, 'up')

    # | == down link
    def __or__(self, i2):
        """ Override the ^ operator to mean to add an 'up' edge """
        return self.link_edge(i2, 'down')

    # >= == right link
    def __ge__(self, i2):
        """ Override the >= operator to mean to add an 'right' edge """
        return self.link_edge(i2, 'right')

    # <= == left link
    def __le__(self, i2):
        """ Override the <= operator to mean to add an 'left' edge """
        return self.link_edge(i2, 'left')

    # -  == horizontal line
    def __sub__(self, i2):
        """ Override the - operator to add a horizontal line """
        return self.link_edge(i2, 'hline')

    # =  == vertical line
    #def __eq__(self, i2):
    #    return self.link_edge(i2, 'vline')

    # >>=  == vertical dotted arrow
    def __irshift__(self, i2):
        return self.link_edge(i2,'vdotted')


class Stack:
    def __init__(self):
        self._stack = []

    def push(self, e):
        #print(f'\tIn push: {self._stack=}')
        self._stack.append(e)

    def pop(self):
        return self._stack.pop()

    def count(self):
        return len(self._stack)

    def current(self):
        return self._stack[-1]

_groups = Stack()

##########################################################################
# Diagram Classes
class ComponentDiagram(Diagram.DiagramBase):
    def __init__(self, helper, filename=None, **attrs):
        Diagram.DiagramBase.__init__(self, helper, filename, **attrs)
        if _groups.count() == 0:
            #print('Initial push:')
            _groups.push(self)

    def inherit(self, attrs):
        a = self.attrs.copy()

        for k in ['shape', 'fillcolor']:
            if k in a:
                a.pop(k)

        a.update(attrs)
        return a

    def Group(self, shape, label, **attrs):
        g=ComponentSubDiagram(parent=self, id=None, shape=shape, label=label, **self.inherit(attrs))
        #self.collection.add_item(g)
        _groups.current().collection.add_item(g)
        return g
    
    def add_new_item(self, shape, label, **kwargs):
        #kwargs.update({'shape':shape})
        #return self.collection.add_new_item(Item, label, **kwargs)
        i = ComponentDiagramItem(label, shape, **self.inherit(kwargs))
        #print('Adding <{}> to <{}>'.format(label,  _groups.current().label))
        return _groups.current().collection.add_item(i)
        #return self.collection.add_item(i)

    # Item functions: Component, Interface, Actor
    def Component(self, label, **attrs):
        return self.add_new_item(Component, label, **attrs)

    def Interface(self, label, **attrs):
        return self.add_new_item(Interface, label, **attrs)

    def Actor(self, label, **attrs):
        return self.add_new_item('Actor', label, **attrs)

    # Group functions: Cloud, Database, Folder, Frame, Node, Package
    def Cloud(self, label=None, **attrs):
        return self.Group(Cloud, label, **attrs)

    def Database(self, label=None, **attrs):
        return self.Group(Database, label, **attrs)

    def Folder(self, label=None, **attrs):
        return self.Group(Folder, label, **attrs)

    def Frame(self, label=None, **attrs):
        return self.Group(Frame, label, **attrs)

    def Node(self, label=None, **attrs):
        return self.Group(Node, label, **attrs)

    def Package(self, label=None, **attrs):
        return self.Group(Package, label, **attrs)

    def Rectangle(self, label=None, **attrs):
        return self.Group(Rectangle, label, **attrs)

#-------------------------------------------------------------------------
class ComponentSubDiagram(ComponentDiagram, ComponentDiagramItem):
    def __init__(self, parent, id, shape, label, **attrs):
        ComponentDiagram.__init__(self, helper=parent.helper, **attrs)
        ComponentDiagramItem.__init__(self, label, shape, **attrs)
        self.attrs['shape'] = shape

        self.parent = parent
        self.id = id
        self.level = self.parent.level + 1

        self.node_attrs = parent.node_attrs.copy()
        self.edge_attrs = parent.edge_attrs.copy()

        # Denotes that this is to be treated as a group
        self.is_group = False

    # Renders this diagram
    def __str__(self):
        if self.is_group:
            a = self.helper.startSubdiagram(self.id, self.label, **self.attrs)
            a += '\n'.join([str(v) for v in list(self.collection.values())])
            a += self.helper.endSubdiagram(self.id, self.label, **self.attrs)
            return a
        else:
            return ComponentDiagramItem.__str__(self)

    def __enter__(self):
        self.is_group = True
        #print(f'Entering <{self.id}>, <{self.label}>')
        _groups.push(self)
        #print('push done')
        return self

    def __exit__(self, *args):
        if self.is_group:
            g= _groups.pop()
            #print(f'Exiting <{g.id}>, <{g.label}>')

    def all_except(self, *args, **kwargs):
        """ Use: 
                all_except(shape=TypeName)  to exclude by type 
                all_except(a,b,c) to exclude items a,b,c
        """
        ret = []
        k_shape = 'shape'
        shape = kwargs.get(k_shape)
        for v in self.all():
            if not v in args:
                argcheck=True
            else:
                argcheck=False

            typecheck=True
            if shape:
                if v.attrs[k_shape] == shape:
                    typecheck=False

            if argcheck and typecheck:
                ret.append(v)

        return ret


##########################################################################
# Context Classes
class ComponentContextPUML(Diagram.Context):
    def __init__(self, filename=None):
        print('Generating!:',filename)
        self.diagram = ComponentDiagram(Helper, filename)

class ComponentContext(Diagram.Context):
    def __init__(self, helper, filename=None):
        _groups._stack=[]
        print (f'Generating--: {filename}')
        self.diagram = ComponentDiagram(helper, filename)
